/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006-2010 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

#include "luav.h"

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>
#include <suplib/str.h>

#include "module.h"
#include "ray_transforms.h"
#include "aperture.h"
#include "common.h"

/* forward declaration of function local to this file */
static int print_ray( lua_State *L );
static void init(Aperture *aperture, void* amip);
static void cleanup(Aperture *aperture);

static ProcessResult process_print_xfrm(Aperture *aperture);
static ProcessResult process_print_raw(Aperture *aperture);
static ProcessResult process_print_raw_project(Aperture *aperture);
static ProcessResult process_print_project(Aperture *aperture);

/*
   Define the ApertureModule structure for this module.
*/
ApertureModule print_ray_module =
{
  "print_ray",	        /* the lua name of the instantiation function */
  print_ray,		/* the C lua-callable instantiation function */
  init,			/* the aperture initialization function */
  NULL,			/* the aperture ray processing function */
  cleanup		/* the aperture cleanup function */
};

/* for backwards compatibility */
ApertureModule print_photon_module =
{
  "print_photon",	/* the lua name of the instantiation function */
  print_ray,		/* the C lua-callable instantiation function */
  init,			/* the aperture initialization function */
  NULL,			/* the aperture ray processing function */
  cleanup		/* the aperture cleanup function */
};

typedef struct {

    ApertureModule module;
    char           *pfx;

} PrintStuff;

/*
   this structure contains the offsets into the data packet of
   the data required by this aperture.  it is initialized by
   the init() function in this file.
*/

static struct
{
  DVector3 *position;
  DVector3 *direction;
} ray;

static int print = 1;

static const char* arglist [] = {
    "xfrm",
    "raw",
    "raw_project",
    "project",
    "on",
    "off",
    NULL
};

enum { XFRM, RAW, RAW_PROJECT, PROJECT, ON, OFF };

/*
   this function is called from lua to create an instantiation of
   this aperture type.
*/
static int
print_ray( lua_State *L ) {

  int option;
  const char* pfx;
  PrintStuff *data;

  tf_enter("print_ray");

  if ( NULL == (data = malloc( sizeof(PrintStuff) )) )
      tf_exit( ExitERR_alloc, "unable to allocate data for print_ray" );

  option = luaL_checkoption( L, 1, NULL, arglist );
  pfx    = luaL_optstring( L, 2, NULL );

  luaL_argcheck( L, 3 > lua_gettop(L), 3, "too many arguments" );

  switch ( option )
  {
  case XFRM:
      print_ray_module.process = process_print_xfrm;
      break;

  case RAW:
      print_ray_module.process = process_print_raw;
      break;

  case RAW_PROJECT:
      print_ray_module.process = process_print_raw_project;
      break;

  case PROJECT:
      print_ray_module.process = process_print_project;
      break;

  case ON:
      print = 1;
      tf_leave();
      return 0;

  case OFF:
      print = 0;
      tf_leave();
      return 0;

  default:
      luaL_error( L, "internal error" );
  }


  data->module = print_ray_module;
  data->pfx = str_dup( pfx );

  aperture_new( &(data->module), data, data->pfx );

  tf_leave();

  return 0;
}

/*

   This routine is called once, before any rays arrive.  It should
   be used to determine offsets to whatever data this aperture needs
   from the ray data packet.  It should also initialize anything
   that needs initialization.

*/

static void
init(Aperture *aperture, void* amip)
{
  ApertureModuleInfo* ami = amip;
  DpktField *dpktf_pos, *dpktf_dir;

  static int initialized = 0;

  /* do this only once to save some CPU */
  if (! initialized)
  {
    tf_enter("print_ray init");

    if ( NULL == (dpktf_pos = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_pos )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_pos
	                           "' in input stream" );

    if ( NULL == (dpktf_dir = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_dir )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_dir
	                           "' in input stream" );

    ray.position  = (DVector3 *) bpipe_dpktf_data( dpktf_pos, ami->dpkt );
    ray.direction = (DVector3 *) bpipe_dpktf_data( dpktf_dir, ami->dpkt );

    initialized = 1;

    tf_leave();
  }
}

/*

   This routine is called for every ray.  It should be as efficient
   as possible.

*/

static void
print_out_ray(
    const char* pfx,
    unsigned long assembly,
    unsigned long aperture,
    DVector3 *p,
    DVector3 *d
    )
{
    if ( pfx )
	tf_message("%s: %03d.%03d: %.14g\t%.14g\t%.14g :\t%.14g\t%.14g\t%.14g\n",
		   pfx, assembly, aperture,
		   p->x, p->y, p->z, d->x, d->y, d->z);
    else
	tf_message("%03d.%03d: %.14g\t%.14g\t%.14g :\t%.14g\t%.14g\t%.14g\n",
		   assembly, aperture,
		   p->x, p->y, p->z, d->x, d->y, d->z);

}

static ProcessResult
process_print_xfrm( Aperture *aperture )
{
  PrintStuff *ps = (PrintStuff *) aperture->data_object;

  DVector3 p;		/* position of vector in module ref frame */
  DVector3 d;		/* direction of vector in module ref frame */

  if (print)
  {
    /* apply transformation to the ray */
    TRANSFORM_POSITION ( aperture->forward, ray.position, &p );
    TRANSFORM_DIRECTION( aperture->forward, ray.direction, &d );

    print_out_ray(ps->pfx, aperture->id.assembly, aperture->id.aperture, &p, &d);
  }

  return NOTAPPLICABLE;
}

static ProcessResult
process_print_project( Aperture *aperture )
{
  PrintStuff *ps = (PrintStuff *) aperture->data_object;

  DVector3 p;		/* position of vector in module ref frame */
  DVector3 d;		/* direction of vector in module ref frame */

  if (print)
  {
    /* apply transformation to the ray */
    TRANSFORM_POSITION ( aperture->forward, ray.position, &p );
    TRANSFORM_DIRECTION( aperture->forward, ray.direction, &d );

    /* translate it to the z = 0 plane */
    PROJECT_RAY(&p, &d, 0.0);

    print_out_ray(ps->pfx, aperture->id.assembly, aperture->id.aperture, &p, &d);
  }

  return NOTAPPLICABLE;
}

static ProcessResult
process_print_raw_project( Aperture *aperture )
{
  PrintStuff *ps = (PrintStuff *) aperture->data_object;

  DVector3 p;		/* position of vector in module ref frame */
  DVector3 d;		/* direction of vector in module ref frame */

  DVector3 p_r;		/* position of vector in module ref frame */
  DVector3 d_r;		/* direction of vector in module ref frame */

  if (print)
  {
    /* apply transformation to the ray */
    TRANSFORM_POSITION ( aperture->forward, ray.position, &p );
    TRANSFORM_DIRECTION( aperture->forward, ray.direction, &d );

    /* translate it to the z = 0 plane */
    PROJECT_RAY(&p, &d, 0.0);

    /* now translate it back to the external frame */
    TRANSFORM_POSITION ( aperture->backward, &p, &p_r );
    TRANSFORM_DIRECTION( aperture->backward, &d, &d_r );

    print_out_ray(ps->pfx, aperture->id.assembly, aperture->id.aperture, &p_r, &d_r);
  }

  return NOTAPPLICABLE;
}

static ProcessResult
process_print_raw( Aperture *aperture )
{
  PrintStuff *ps = (PrintStuff *) aperture->data_object;

  if (print)
    print_out_ray(ps->pfx, aperture->id.assembly, aperture->id.aperture,
		 ray.position, ray.direction);

  return NOTAPPLICABLE;
}


/*
   This routine is called once for every instantiatian of this aperture.
   It should free instantiation specific resources (like those allocated
   in the lua-called instantiation creation routine).
*/

static void
cleanup (Aperture *aperture)
{
    PrintStuff *ps = (PrintStuff *) aperture->data_object;

    free( ps->pfx );

    free( ps );
}








