/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006-2010 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>

#include "luav.h"

#include <bpipe/bpipe.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <mst_nrutil/mst_nrutil.h>

#include "lua_register.h"
#include "aperture.h"

#include "xmatrix.h"
#include "xstack.h"

#include "common.h"

static const char* const XStr[2] = { "intrinsic", "extrinsic" };

/*---------------------------------------------------------------------*/
/*
  PROCEDURE
    Rotation of Co-ordinate Axes.

  DESCRIPTION
    Calculate the 3-D axes rotation, the return pointer of the
    procedure.  The axes are rotated anti-clockwise through an angle
    theta radians about an  axis specified by the argument `axis'.
    If `axis' is XAXIS, YAXIS or ZAXIS then the rotation shall be about
    the axis x, y or z, respectively.

  AUTHOR
    Dan Nguyen
*/
static void
  rotate(
	 XYZaxis axis,		/* axis of rotation */
	 double  angle,		/* angle to rotate, in radians */
	 CoordTransform coord_xfrm
	 )
{
  XMatrix mat;

  tf_enter("rotate");

  mat = XMatrix_rotation( axis, angle );
  xstack_update( mat, coord_xfrm );
  XMatrix_free( mat );

  tf_leave();
  return;
}

/*---------------------------------------------------------------------*/
/*
  PROCEDURE
    Rotation about X-Co-ordinate axis.

  DESCRIPTION
    Calculate the 3-D axis rotation about the x-axis of `angle' radian.
    The orientation of the rotation is anti-clockwise.

  AUTHOR
    Dan Nguyen
*/
static int
rotate_x_lua( lua_State *L )
{
  double angle;
  tf_enter("rotate_x");

  angle = luaL_checknumber( L, 1 );
  luaL_argcheck( L, 2 >= lua_gettop(L), 3, "too many arguments" );

  rotate( XAXIS, angle, luaL_checkoption( L, 2, XStr[intrinsic], XStr ) );

  tf_leave();

  return 0;
}

/*
  PROCEDURE
    Rotation about Y-Co-ordinate axis.

  DESCRIPTION
    Calculate the 3-D axis rotation about the y-axis of `angle' radian.
    The orientation of the rotation is anti-clockwise.

  AUTHOR
    Dan Nguyen
*/
static int
rotate_y_lua( lua_State *L )
{
  double angle;
  tf_enter("rotate_y");

  angle = luaL_checknumber( L, 1 );
  luaL_argcheck( L, 2 >= lua_gettop(L), 3, "too many arguments" );

  rotate( YAXIS, angle, luaL_checkoption( L, 2, XStr[intrinsic], XStr ) );

  tf_leave();

  return 0;
}

/*
  PROCEDURE
    Rotation about Z-Co-ordinate axis.

  DESCRIPTION
    Calculate the 3-D axis rotation about the z-axis of `angle' radian.
    The orientation of the rotation is anti-clockwise.

  AUTHOR
    Dan Nguyen
*/
static int
rotate_z_lua( lua_State *L )
{
  double angle;
  tf_enter("rotate_z");

  angle = luaL_checknumber( L, 1 );
  luaL_argcheck( L, 2 >= lua_gettop(L), 3, "too many arguments" );

  rotate( ZAXIS, angle, luaL_checkoption( L, 2, XStr[intrinsic], XStr ) );

  tf_leave();
  return 0;
}


/*

  PROCEDURE

    check if theres an NxN table on the stack

  AUTHOR
    Diab Jerius

*/

static XMatrix
load_matrix( lua_State *L, int idx, int N ) {

    XMatrix mat = NULL;

    int top = lua_gettop( L ) ;

    int row, col;

    /* check for table and number of rows */
    if ( ! lua_istable( L, idx ) || lua_objlen(L, idx ) != N )
	return NULL;

    mat = XMatrix_identity();
    /* iterate through rows of matrix */
    for ( row = 0 ; row < N ; ++row ) {

	lua_pushinteger( L, row+1 );
	lua_gettable( L, idx );

	/* check for number of columns */
	if ( lua_objlen(L, -1) != N )
	    goto error;

	/* check that elements are numbers */
	for ( col = 0 ; col < N ; ++col ) {

	    lua_pushinteger( L, col+1 );
	    lua_gettable( L, -2 );

	    if ( ! lua_isnumber( L, -1 ) )
		goto error;

	    mat[row][col] = lua_tonumber( L, -1 );
	    lua_pop( L, 1 );
	}

	lua_pop( L, 1 );
    }

    goto success;

error:

    if ( mat != NULL ) {
	XMatrix_free( mat );
	mat = NULL;
    }


success:

    lua_settop( L, top );
    return mat;
}


/*
  PROCEDURE
    Rotation with a passed 3x3 matrix.

  DESCRIPTION

  AUTHOR
    Diab Jerius
*/
static int
rotate_lua( lua_State *L )
{
  XMatrix mat;

  luaL_argcheck( L, 2 >= lua_gettop(L), 3, "too many arguments" );

  if ( NULL == (mat = load_matrix( L, 1, 3 )) )
      luaL_argerror( L, 1, "not a 3x3 matrix of numbers" );

  xstack_update( mat, luaL_checkoption( L, 2, XStr[intrinsic], XStr ) );
  XMatrix_free( mat );

  return 0;
}


/*
   a lua function to translate the matrix
*/
static int
translate_lua( lua_State *L )
{
  XMatrix mat;
  double transl[DIM3];

  tf_enter("translate");

  transl[0] = luaL_checknumber( L, 1 );
  transl[1] = luaL_checknumber( L, 2 );
  transl[2] = luaL_checknumber( L, 3 );

  luaL_argcheck( L, 3 == lua_gettop(L) || 4 ==lua_gettop(L), 5,
		 "too many arguments" );



  mat = XMatrix_translation( transl );
  xstack_update( mat, luaL_checkoption( L, 4, XStr[intrinsic], XStr ) );
  XMatrix_free( mat );

  tf_leave();
  return 0;
}

/*
  PROCEDURE
    Transformation with a passed 4x4 matrix.

  DESCRIPTION

  AUTHOR
    Diab Jerius
*/
static int
transform_lua( lua_State *L )
{
  XMatrix mat;

  luaL_argcheck( L, 2 >= lua_gettop(L), 3, "too many arguments" );

  if ( NULL == (mat = load_matrix( L, 1, 4 )) )
      luaL_argerror( L, 1, "not a 4x4 matrix of numbers" );

  xstack_update( mat, luaL_checkoption( L, 2, XStr[intrinsic], XStr ) );
  XMatrix_free( mat );

  return 0;
}



/*---------------------------------------------------------------------*/

void
transform_register_lua( SLinkList liblist )
{
    register_luacfunc( liblist, "translate",  translate_lua );
    register_luacfunc( liblist, "rotate_x",   rotate_x_lua  );
    register_luacfunc( liblist, "rotate_y",   rotate_y_lua  );
    register_luacfunc( liblist, "rotate_z",   rotate_z_lua  );
    register_luacfunc( liblist, "rotate",     rotate_lua  );
    register_luacfunc( liblist, "transform",  transform_lua  );
}

/*---------------------------------------------------------------------*/

#if 0
void
  transform_position( XMatrix tran, DVector3 *in, DVector3 *out )
{
  out->x = in->x * tran[0][0] + in->y * tran[1][0] + in->z * tran[2][0] +
    tran[3][0];

  out->y = in->x * tran[0][1] + in->y * tran[1][1] + in->z * tran[2][1] +
    tran[3][1];

  out->z = in->x * tran[0][2] + in->y * tran[1][2] + in->z * tran[2][2] +
    tran[3][2];

  return;
}

void
  transform_direction( XMatrix tran, DVector3 *in, DVector3 *out )
{
  out->x = in->x * tran[0][0] + in->y * tran[1][0] + in->z * tran[2][0] ;

  out->y = in->x * tran[0][1] + in->y * tran[1][1] + in->z * tran[2][1] ;

  out->z = in->x * tran[0][2] + in->y * tran[1][2] + in->z * tran[2][2] ;

  return;
}
#endif


