/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006-2010 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "luav.h"

#include <linklist/linklist.h>

#include "xmatrix.h"

#include "init.h"
#include "lua_register.h"
#include "xstack.h"

#include "transform.h"

#include "common.h"

/* the transformation stack.  it's empty inbetween processing assemblies */
static SLinkList xstack = NULL;

/*---------------------------------------------------------------------*/
/*
   PROCEDURE
   xstack_init - initialize global structures before assemblies are
   built
*/
void
initialize_xstack( void )
{
  tf_enter("xstack_init");

  /* allocate transformation stack */
  tf_assert(xstack == NULL);

  if ( NULL == (xstack = sll_new( NULL )) )
    tf_die( "error creating transformation matrix stack" );

  if ( sll_insert_head( xstack, XMatrix_identity( ) ) )
    tf_exit( ExitERR_alloc, "unable to insert global coord matrix in stack");

  tf_leave();
}

void
cleanup_xstack( void )
{
  sll_delete( xstack, (SLL_DELETE) XMatrix_free );
}

static XMatrix
xstack_top( void )
{
  XMatrix stack_top;

  stack_top = sll_head( xstack );
  tf_assert(stack_top != NULL);

  return stack_top;
}

/* return a copy of the top xfrm matrix */
XMatrix
xstack_top_copy( void )
{
  XMatrix stack_top;

  stack_top = sll_head( xstack );
  tf_assert(stack_top != NULL);

  return XMatrix_dup( stack_top );
}


/*

  PROCEDURE
  xstack_push

  DESCRIPTION
  Push a transformation matrix onto the stack. The caller must
  not free the passed matrix.

  RETURNS
  void

*/
void
xstack_push( XMatrix mat )
{
  tf_enter( "xstack_push" );

  if ( sll_insert_head( xstack, mat ) )
    tf_die( "couldn't insert node into stack" );

  tf_leave( );
}



/*

  PROCEDURE
  xstack_pop

  DESCRIPTION
  Pop off the top matrix from the transformation stack.  The value of the
  parameter empty_is_fatal determines how it reacts to an empty stack.
  If empty_is_fatal is true, an empty stack causes the program to exit.
  If it is false, an empty stack causes it to return non-zero.  If the
  stack isn't empty, it'll return zero.

  Note that an "empty" stack really has one element on it, the initial
  matrix pushed there by xstack_init.  This should, of course only
  be removed by cleanup_aperture.  Thus, an empty stack ain't empty.

*/
int
xstack_pop(
  int empty_is_fatal	/* if true, an empty stack is a fatal error,
			   else, return non-zero if empty stack */
)
{
  SLLNode head_node;
  XMatrix stack_top;

  tf_enter("xstack_pop");

  head_node = sll_head_node( xstack );
  tf_assert( head_node != NULL );

  if ( sll_count( xstack ) <= 1 )
  {
    if ( empty_is_fatal )
      tf_assert( sll_count(xstack) <= 1 );
    else
      return 1;
  }

  stack_top = sll_destroy_node( xstack, head_node );
  tf_assert( stack_top != NULL );

  XMatrix_free( stack_top );

  tf_leave();

  return 0;
}


/*
  PROCEDURE
    Calculate the CTM 

  DESCRIPTION
    Multiply all of the elements on the stack to generate the CTM.

  RETURNS
    an XMatrix
*/

XMatrix
xstack_get_CTM( void  )
{
  SLLNode node;
  int i;
  int idx;
  XMatrix ctm[2];

  if ( ! xstack )
      return NULL;

  ctm[0] = XMatrix_identity();
  ctm[1] = XMatrix_new();

  /* stack is in LIFO order, same order of the matrix product,
     so rightmultiply the matrices */

  for ( i = 0, node = sll_head_node( xstack ), idx = 0;
	 node ;
	i++, node = sll_next_node( xstack, node ), idx = 1 - idx ) {

      XMatrix mat = sll_node_get_data( node );
      tf_assert(mat != NULL);

      XMatrix_mult( ctm[idx], mat, ctm[1-idx] );
  }

  XMatrix_free( ctm[1-idx] );

  return ctm[idx];
}



/*
  PROCEDURE
    Update the top element of the stack.

  DESCRIPTION
    Pop the top element of the stack.
    Multiply the element which was popped off with the incoming matrix.
    Store the resulting matrix as the top element of the stack.

  RETURNS
   void
*/
void xstack_update( XMatrix mat, CoordTransform xfrm_type )
{
  SLLNode head_node;
  XMatrix stack_top;
  XMatrix top;

  tf_enter("xstack_update");

  /*
     get the matrix from the top of the stack.  leave the node on the
     xstack, we'll just update it later
  */
  head_node = sll_head_node( xstack );
  tf_assert(head_node != NULL);

  stack_top = sll_node_get_data( head_node );
  tf_assert(stack_top != NULL);

  tf_assert( xfrm_type == intrinsic || xfrm_type == extrinsic );

  top = xfrm_type == intrinsic
      ? XMatrix_mult( mat, stack_top, NULL )
      : XMatrix_mult( stack_top, mat, NULL );

  /* now update the data in the node at the top of the stack */
  sll_node_put_data( head_node, top );

  /* free the old top of stack matrix */
  XMatrix_free( stack_top );

  tf_leave();
}



/*
  PROCEDURE
  xstack_check

  DESCRIPTION
  This routine is called after the lua script has been executed.
  If there are still entries on the transformation stack, there
  were too many push_CTM's, or not enough pop_CTM's performed
*/
void
xstack_check( void )
{
  /* there should be one CTM left on the stack when all is done */
  if ( sll_count( xstack ) > 1 )
    tf_die( "unbalanced push_CTM/pop_CTM calls: too few pop_CTM calls" );
}

void
  xstack_print_stack(  )
{
  XMatrix mat;
  SLLNode node;
  int i;

  if ( xstack )
    for ( i = 0, node = sll_head_node( xstack );
	 node ;
	 i++, node = sll_next_node( xstack, node ) ) {
      mat = sll_node_get_data( node );
      tf_assert(mat != NULL);
      tf_message( "Transformation stack, i = %d\n", i );
      XMatrix_print( mat );
    }
  return;
}

static int
xstack_print_stack_lua( lua_State *L )
{
    luaL_argcheck( L, 0 == lua_gettop(L), 1, "too many arguments" );
    xstack_print_stack( );
    return 0;
}

void
xstack_print_top ( int invert )
{
  if ( xstack )
  {
    tf_message("Transformation matrix:\n");

    if ( invert ) {

	XMatrix inv = XMatrix_inverse( xstack_top() );
	XMatrix_print( inv ) ;
	XMatrix_free( inv );

    }

    else
	XMatrix_print(xstack_top());

  }
  else
    tf_message("no current transformation\n");
}

static int
xstack_print_CTM_lua ( lua_State *L )
{
    XMatrix mat = xstack_get_CTM();
    XMatrix inv = XMatrix_inverse( mat );

    luaL_argcheck( L, 0 == lua_gettop(L), 1, "too many arguments" );

    tf_message( "Transformation matrix:\n" );
    tf_message( "Forward:\n" );
    XMatrix_print( inv );
    tf_message( "Backward:\n" );
    XMatrix_print( mat );

    XMatrix_free( mat );
    XMatrix_free( inv );

    return 0;
}

static int
xstack_print_stack_top_lua ( lua_State *L )
{
    luaL_argcheck( L, 0 == lua_gettop(L), 1, "too many arguments" );
    xstack_print_top( 0 );

    return 0;
}

static int
xstack_get_top_lua( lua_State *L )
{
    int i, j;
    XMatrix top;
    int invert;

    if ( ! xstack ) {
	lua_pushnil( L );
	return 1;
    }

    invert = lua_toboolean( L, 1 );
    top = invert ? XMatrix_inverse( xstack_top() ) : xstack_top();

    lua_createtable( L, 4, 0 );

    for ( i = 0; i < DIM4; i++ ) {
	lua_pushinteger( L, i+1 );
	lua_createtable( L, 4, 0 );

	for ( j = 0; j < DIM4; j++ ) {

	    lua_pushinteger( L, j+1);
	    lua_pushnumber( L, top[i][j] );
	    lua_settable( L, -3 );
	}
	lua_settable( L, -3 );
    }

    if ( invert )
	XMatrix_free( top );

    return 1;
}


/*---------------------------------------------------------------------*/

void
xstack_register_lua( SLinkList liblist )
{
    register_luacfunc( liblist, "print_CTM",      xstack_print_CTM_lua );
    register_luacfunc( liblist, "get_TM_stack_top",        xstack_get_top_lua );
    register_luacfunc( liblist, "print_TM_stack", xstack_print_stack_lua );
    register_luacfunc( liblist, "print_TM_stack_top", xstack_print_stack_top_lua );
}

