// --8<--8<--8<--8<--
//
// Copyright (C) 2013 Smithsonian Astrophysical Observatory
//
// This file is part of bp-save-history
//
// bp-save-history is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// -->8-->8-->8-->8--

#include <string>
#include <cstdlib>

#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>

using namespace boost;

#include <bpipe/bpipe.h>
#include <Exception/Exception.h>

#include "field.h"

Field::Field( BPipe* bpipe, std::string& name ) {

    dpktf = bpipe_dpktf( bpipe, name.c_str() );
    if ( dpktf == NULL )
	throw( Exception( "field " + name + " does not exist" ) );

    _init();
}

Field::Field( DpktField* dpktf ) : dpktf(dpktf) {

    _init();

}

void
Field::set_data( void* core_image ) {

    data = bpipe_dpktf_data( dpktf, core_image );
    src = data + copy_matrix->ne * datatype_size;

}

void
Field::copy_data( ) {

    // data are stored as 1D array of history, where
    // data[0] is the youngest.  at this point in processing,
    // data has been extended so data[0] is uninitialized. this
    // routine copies data[1] to data[0].

    // this should really be performed by a bpipe routine.  but
    // that's not available
    bpipe_data_copy( data, src, copy_matrix.get(), datatype );

}

void
Field::_init( ) {

    datatype = bpipe_dpktf_type( dpktf );
    datatype_size = bpipe_datatype_size( datatype );

    // this describes the shape of the data which must be replicated
    // in the extended field. at the moment only support scalar data
    copy_matrix = shared_ptr<BPMatrix>( bpipe_matrix_new_va( 1, 1 ), free );

    _extend();
}

void
Field::_extend() {

    BPMatrix *bpmatrix;

    /* get matrix description for core image */
    if ( NULL == ( bpmatrix = bpipe_dpktf_matrix ( dpktf, BPDSite_CORE, NULL )))
	throw( Exception( str( format( "error extending field %1%: %2%" )
			       % bpipe_dpktf_name( dpktf )
			       % bpipe_strerror(bpipe_errno))
			  ) );

    if ( bpmatrix->nd != 1 )
	throw( Exception( str( format( "error extending field %1%: cannot handle multi-dimensional fields"  )
				       % bpipe_dpktf_name( dpktf ) ) ) );

    /* ... and now extend it */
    ++ bpmatrix->extent[0];

    //--------------------------------------------------------
    // resize core image such that the new slot is at the origin of
    // the structure

    // this is freed by the bpipe library
    size_t *dst_off = bpipe_offset_new_va( 1, 1 );

    if (bpipe_dpktf_resize_core( dpktf, bpmatrix, NULL, dst_off, NULL))
	throw( Exception( str( format( "error extending field %1%: %2%" )
			       % bpipe_dpktf_name( dpktf )
			       % bpipe_strerror(bpipe_errno)) ) );
}

list<Field>
gen_fields( BPipe* bpipe, list<string>& field_names ) {

    list<Field> fields;

    if ( field_names.size() == 1 && field_names.front() == "%all" ) {

	void *p_last = NULL;
	DpktField *dpktf;

	while( ( dpktf = bpipe_dpktf_next( bpipe, &p_last ) ) ) {

	    fields.push_back( Field( dpktf ) );
	}
    }

    else {

	list<string>::iterator name;

	for( name = field_names.begin() ; name != field_names.end() ; ++name ) {

	    fields.push_back( Field( bpipe, *name ) );

	}
    }

    return fields;
}
