// --8<--8<--8<--8<--
//
// Copyright (C) 2013 Smithsonian Astrophysical Observatory
//
// This file is part of bpipexx
//
// bpipexx is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// -->8-->8-->8-->8--

#include <iostream>
#include <string>
#include <vector>

#include "Exception/Exception.h"

#include "bpipexx.h"

using namespace std;

/**

   \class SAOTrace::BPipeX::Simple
   \headerfile bpipexx/bpipexx.h

   A simple wrapper class providing optimized access to a binary pipe
   object.

   \section simple_open_streams Creating an object and attaching to streams

   The easiest means to create an object manipulating bpipe streams is to
   use the two argument constructor:

   \code

   #include <SAOTrace/bpipexx/bpipexx.h>
   using namespace SAOTrace;

   BPipeX::Simple bp( "", "stdout" );

   \endcode

   In this case it opens an output stream, but no input stream,
   suitable for a ray generater.  For a filter, specify both streams,
   for a sink, specify only the input stream.

   For more complex situations, don't specify any streams, and use the
   \c bpipe accessor to manipulate the underlying \c bpipe object
   directly:

   \code
   BPipeX::Simple bp();
   bpipe_input( bp->bpipe, "stdin" );
   \endcode

   \section simple_manip_hdrf  Manipulating Header Fields

   Recall that the header accepts multiple fields with the same name.

   The first step is to create a \c BPipeX::HdrF object, then pass it
   to the \c add method.  The steps can be combined:

   \code
   bp.add( BPipeX::HdrF<double>("temp", 33.2 ) );
   \endcode

   To delete a header field, use \c delete_hdrf:

   \code
   bp.delete_hdrf( "temp" );
   \endcode

   \section simple_manip_dpktf  Manipulating Data Packet Fields

   Recall that the data packets can have only one field with a given
   name.

   The first step is to create a \c BPipeX::DpktF object, then pass it
   to the \c add method.  The steps can be combined:

   \code
   bp.add( BPipeX::DpktF<double>("temp") );
   \endcode

   To delete a data packet field from the output, use \c delete_dpktf:

   \code
   bp.delete_dpktf( "temp" );
   \endcode

   See the documentation for \c delete_dpktf for more complex use.

   \section simple_map_read_write Mapping, reading and writing

   After all fields are created, the packet images must be mapped:

   \code
   bp.map();
   \endcode

   To access the location in the core image for a data packet
   field, use  \c dpktf_data:

   \code
   double* temp_p = bp.dpktf_data<double>( "temp" );
   \endcode

   For filters or sinks, make sure to write the header prior to
   outputing any data packets:

   \code
   bp.write_hdr();
   \endcode

   Use \c read_dpkts and \c write_dpkts to read and write.

   \section simple_close_streams Closing I/O Streams

   The streams will be closed upon destruction of the \c
   BPipeX::Simple object.


**/

namespace SAOTrace {

    namespace BPipeX {


	/** \brief Constructor specifying input and output streams
	 *
	 * Either of the parameters may be an empty string, indicating
	 * that that stream should not be attached.
	 *
	 * \param[in] input the name of the input stream (may be \c stdin)
	 * \param[in] output  the name of the input stream (may be \c stdout)
	 * \throws Exception if it is unable to attach the streams.
	 *
	 */
	Simple::Simple( const string& input, const string& output ) {

	    _bpipe = bpipe_new();

	    if ( input.size() ) {

		if ( bpipe_input( _bpipe, input.c_str() ) )
		    throw Exception( bpipe_strerror( bpipe_errno ) );
	    }

	    if ( output.size() ) {

		BPipeOutput* bpo_p = bpipe_output( _bpipe, output.c_str() );

		if ( NULL == bpo_p )
		    throw Exception( bpipe_strerror( bpipe_errno ) );

		bpo.push_back( bpo_p );
	    }
	}


	/** \brief Constructor
	 *
	 * Construct the underlying \c bpipe object without attaching
	 * any streams
	 *
	 **/
	Simple::Simple( ) {
	    _bpipe = bpipe_new() ;
	}

	/** \brief Destructor
	 *
	 * Destroy a BPipeX::Simple object, closing the associated
	 * output streams.
	 *
	 **/
	Simple::~Simple( ) {
	    bpipe_delete( _bpipe ) ;
	}

	/** \brief map the data packet fields onto memory.
	 *
	 * This function maps the data packet fields onto memory and allocates
	 * space for the images.
	 *
	 * \param[in] npkts the \e optional number of packets to allocate space for. defaults to 1
	 * \throws Exception if it is unable to allocate the image.
	 *
	 */
	void
	Simple::map( int npkts ) {

	    n_pkts = npkts;
	    core_image = bpipe_map_alloc( _bpipe, n_pkts, &core_image_size );

	    if ( NULL == core_image )
		throw Exception( bpipe_strerror( bpipe_errno ) );

	}

	/** \brief read in the next batch of data packets
	 *
	 * Read in the number of data packets specified by the \c map method.
	 *
	 * \returns the number of packets read
	 * \throws Exception if there is an error reading the input stream
	 *
	 */
	size_t
	Simple::read_dpkts () {

	    size_t n = bpipe_read_dpkts( _bpipe, core_image, n_pkts );

	    if ( 0 == n && bpipe_errno == BPEIOERR )
		throw Exception( bpipe_strerror( bpipe_errno ) );

	    return n;
	}

	/** \brief write the output header
	 *
	 * Writes header information to the output channels
	 *
	 * \throws Exception if there is an error
	 *
	 */
	void
	Simple::write_hdr () {

	    if ( bpipe_write_hdr( _bpipe ) )
		throw Exception( bpipe_strerror( bpipe_errno ) );
	}

	/** \brief write a data packet
	 *
	 * Writes a data packet to one or all output channels.
	 *
	 * \param[in] bpo the \e optional output channel to write
	 *                to. Defaults to writing to all output channels.
	 * \throws Exception if there is an error
	 *
	 */
	void
	Simple::write_dpkts( BPipeOutput* bpo ) {

	    if ( bpipe_write_dpkt_n( _bpipe, core_image, n_pkts, bpo ) )
		throw Exception( bpipe_strerror( bpipe_errno ) );

	}

	/** \brief test if a header field exists
	 *
	 *
	 * \param[in] name the name of the field
	 * \param[in] index the \e optional index of the field to
	 *            delete. Defaults to \c BPHdrfIdx_LAST
	 * \returns \c true if the field exists
	 *
	 */
	bool
	Simple::has_hdrf( const char* name, size_t index ) {

	    return bpipe_hdrf_data( _bpipe, name, index ) != NULL;
	}


	/** \brief test if a data packet field exists
	 *
	 *
	 * \param[in] name the name of the field
	 * \returns \c true if the field exists
	 *
	 */
	bool
	Simple::has_dpktf( const char* name ) {

	    return bpipe_dpktf( _bpipe, name ) != NULL;
	}


	/** \brief delete a header field
	 *
	 * Deletes the specified header field.  By default it will delete
	 * all header fields with the given name.  Set the \p index parameter
	 * to either the index of field to delete, or to \c BPHdrfIdx_LAST
	 * to delete the last one.
	 *
	 * \param[in] name the name of the field to delete
	 * \param[in] index the \e optional index of the field to delete.
	 * \throws Exception if there is an error
	 *
	 */
	void
	Simple::delete_hdrf( const char* name, size_t index ) {

	    if ( bpipe_hdrf_delete( _bpipe, name, index ) && bpipe_errno )
		throw Exception( bpipe_strerror( bpipe_errno ) );
	}

	/** \brief delete a data packet field
	 *
	 * Deletes the specified data packet field.  By default the
	 * field is removed from the output image of all output
	 * channels.  Set the \p site parameter to change the data site
	 * or the \p channel parameter to specify an alternate output channel
	 *
	 * \param[in] name the name of the field to delete
	 * \param[in] site the \e optional data site in which to delete it.  This
	 *            may be one of
	 *            \arg \c BPDSite_INPUT
	 *            \arg \c BPDSite_CORE, or
	 *            \arg \c BPDSite_OUTPUT
	 * \param[in] channel \e optional channel output channel to delete from
	 * \throws Exception if there is an error
	 *
	 */
	void
	Simple::delete_dpktf( const char* name, BPDataSite site, BPipeOutput *channel ) {

	    if ( bpipe_dpktf_delete( _bpipe, name, site, channel ) < 0 )
		throw Exception( bpipe_strerror( bpipe_errno ) );

	}


	/**
	   \cond IGNORE
	 **/

	template<>
	HdrF<double>::HdrF ( const char* cname, double idatum ) :
	    name(cname), matrix(NULL),  BPT(BPDType_double), datum(idatum), copy(true) {}

	template<>
	HdrF<int>::HdrF ( const char* cname, int idatum ) :
	    name(cname), matrix(NULL),  BPT(BPDType_int), datum(idatum), copy(true) {}

	template<>
	HdrF<unsigned int>::HdrF( const char* cname, unsigned int idatum ) :
	    name(cname), matrix(NULL),  BPT(BPDType_uint), datum(idatum), copy(true) {}

	template<>
	HdrF<DVector2>::HdrF( const char* cname, DVector2 idatum ) :
	    name(cname), matrix(NULL),  BPT(BPDType_DVector2), datum(idatum), copy(true) {}

	template<>
	HdrF<DVector3>::HdrF( const char* cname, DVector3 idatum ) :
	    name(cname), matrix(NULL),  BPT(BPDType_DVector3), datum(idatum), copy(true) {}

	template<>
	HdrF<IVector2>::HdrF( const char* cname, IVector2 idatum ) :
	    name(cname), matrix(NULL),  BPT(BPDType_IVector2), datum(idatum), copy(true) {}

	template<>
	HdrF<IVector3>::HdrF( const char* cname, IVector3 idatum ) :
	    name(cname), matrix(NULL),  BPT(BPDType_IVector3), datum(idatum), copy(true) {}

	template<>
	HdrF<UIVector2>::HdrF( const char* cname, UIVector2 idatum ) :
	    name(cname), matrix(NULL),  BPT(BPDType_UIVector2), datum(idatum), copy(true) {}

	template<>
	HdrF<UIVector3>::HdrF( const char* cname, UIVector3 idatum ) :
	    name(cname), matrix(NULL),  BPT(BPDType_UIVector3), datum(idatum), copy(true) {}

	template<>
	DpktF<double>::DpktF( const char* cname ):
	    name(cname), matrix(NULL), BPT(BPDType_double) {}

	template<>
	DpktF<int>::DpktF( const char* cname ):
	    name(cname), matrix(NULL), BPT(BPDType_int) {}

	template<>
	DpktF<unsigned int>::DpktF( const char* cname ):
	    name(cname), matrix(NULL), BPT(BPDType_uint) {}

	template<>
	DpktF<DVector2>::DpktF( const char* cname ):
	    name(cname), matrix(NULL), BPT(BPDType_DVector2) {}

	template<>
	DpktF<DVector3>::DpktF( const char* cname ):
	    name(cname), matrix(NULL), BPT(BPDType_DVector3) {}

	template<>
	DpktF<IVector2>::DpktF( const char* cname ):
	    name(cname), matrix(NULL), BPT(BPDType_IVector2) {}

	template<>
	DpktF<IVector3>::DpktF( const char* cname ):
	    name(cname), matrix(NULL), BPT(BPDType_IVector3) {}

	template<>
	DpktF<UIVector2>::DpktF( const char* cname ):
	    name(cname), matrix(NULL), BPT(BPDType_UIVector2) {}

	template<>
	DpktF<UIVector3>::DpktF( const char* cname ):
	    name(cname), matrix(NULL), BPT(BPDType_UIVector3) {}

	/**
	    \endcond
	**/

    }

}
