module( ..., package.seeall )

require 'strict'
require 'package'
require 'deepcompare'

va = require('validate.args')

local raygen = require('saotrace.raygen')
local intrinsic = raygen.intrinsic
local coords = require('saotrace.raygen.coords')
local chandra_coords = require('chandra.saotrace.coords')
local string = require('string')
local srcdir = os.getenv('srcdir') or '.'
local ea_db = table.concat( { srcdir, 'data', 'annulus.rdb'}, '/' )
local units   =  require('saotrace.raygen.units')
local convert = units.convert
local argfmt = string.format( '%%s; file = %q; shell = %d', ea_db, 1 )

local min2rad = coords.min2rad
local deg2rad = coords.deg2rad

function setup()

   _G.setup()

   -- saotrace.config gets reset in _G.setup(), so load the new one
   config = require('saotrace.config').raygen

   -- need to properly reload this for each test to get
   -- clean environment
   package.loaded['chandra.saotrace.raygen.focus'] = nil
   package.loaded['chandra.saotrace.raygen.source'] = nil

end


function test_focus_default()

   config.scripts.entrance_aperture.args = { file = ea_db, shell = 1 }
   config.scripts.sources.args = { spectrum = 1.49 }
   raygen.init( 'chandra.saotrace.raygen.focus' )

   local stash = require('saotrace.raygen.intrinsic').stash

   -- start the source
   assert_true( deepcompare( { { name = 'point1'},
			       func = 'start',
			       mod = 'source'
			    },
			     stash[1]),
		'start source' )

   -- the order of the following is dependent upon what happens in
   -- saotrace.raygen.source.create

   assert_equal( 'extent', stash[2].mod );
   assert_equal( 'point',    stash[2].func );
   assert_equal( 0,    stash[2][1].position.theta[1] );
   assert_equal( 0,    stash[2][1].position.phi[1] );

   assert_equal( 'spectrum', stash[3].mod );
   assert_equal( 'mono',     stash[3].func );
   assert_equal( 1.49,       stash[3][1].energy );
   assert_equal( 1,          stash[3][1].flux );

   -- finish the source
   assert_true( deepcompare( { { name = 'point1' },
			       func = 'finish',
			       mod = 'source'
			    },
			     stash[4]),
		'finish source' )


   -- check the entrance aperture
   assert_equal( 'ea',       stash[#stash].mod );
   assert_equal( 'annulus',  stash[#stash].func );
   assert_equal( 10,         stash[#stash][1].opts.ro );

end

function test_focus_spectrum()

   raygen.init( 'chandra.saotrace.raygen.focus',
		string.format( argfmt, 'spectrum = 44')
	     )

   local stash = require('saotrace.raygen.intrinsic').stash

   -- start the source
   assert_true( deepcompare( { { name = 'point1'},
			       func = 'start',
			       mod = 'source'
			    },
			     stash[1]),
		'start source' )

   -- the order of the following is dependent upon what happens in
   -- saotrace.raygen.source.create

   assert_equal( 'extent', stash[2].mod );
   assert_equal( 'point',    stash[2].func );
   assert_equal( 0,    stash[2][1].position.theta[1] );
   assert_equal( 0,    stash[2][1].position.phi[1] );

   assert_equal( 'spectrum', stash[3].mod );
   assert_equal( 'mono',     stash[3].func );
   assert_equal( 44,         stash[3][1].energy );
   assert_equal( 1,          stash[3][1].flux );

   -- finish the source
   assert_true( deepcompare( { { name = 'point1' },
			       func = 'finish',
			       mod = 'source' },
			     stash[4]),
		'finish source' )

   -- check the entrance aperture
   assert_equal( 'ea',       stash[#stash].mod );
   assert_equal( 'annulus',  stash[#stash].func );
   assert_equal( 10,         stash[#stash][1].opts.ro );


end

function test_focus_position()

   local theta, phi = 20, 3

   raygen.init( 'chandra.saotrace.raygen.focus',
		string.format( argfmt, 'position = { theta = '
			       .. theta .. ', phi = '
			       .. phi .. ' }; spectrum = 1.49')
	     )

   theta, phi = chandra_coords.MSC_to_raygen_polar( theta * min2rad , phi * deg2rad )
   local stash = require('saotrace.raygen.intrinsic').stash

   -- start the source
   assert_true( deepcompare( { { name = 'point1'},
			       func = 'start',
			       mod = 'source'
			    },
			     stash[1]),
		'start source' )

   -- the order of the following is dependent upon what happens in
   -- saotrace.raygen.source.create

   assert_equal( 'extent', stash[2].mod );
   assert_equal( 'point',    stash[2].func );
   assert_equal( theta, convert( stash[2][1].position.theta, nil, units.angular.radian)[1] );
   assert_equal( phi,   convert( stash[2][1].position.phi, nil, units.angular.radian)[1] );

   assert_equal( 'spectrum', stash[3].mod );
   assert_equal( 'mono',     stash[3].func );
   assert_equal( 1.49,       stash[3][1].energy );
   assert_equal( 1,          stash[3][1].flux );

   -- finish the source
   assert_true( deepcompare( { { name = 'point1' },
			       func = 'finish',
			       mod = 'source' },
			     stash[4]),
		'finish source' )

   -- check the entrance aperture
   assert_equal( 'ea',       stash[#stash].mod );
   assert_equal( 'annulus',  stash[#stash].func );
   assert_equal( 10,         stash[#stash][1].opts.ro );


end
