#include <iostream>
#include <string>

using namespace std;

#include <cxcparam/parameter.h>

#include <Exception/Exception.h>

#include <cxcparamx/simple.h>
#include <cxcparamx/psimple.h>

/**
    \class cxcparamx::simple
    \headerfile cxcparamx/simple.h

    A simple wrapper class providing polymorphic attribute retrieval.

    Instantiating the class opens the parameter file; the file is
    closed during object destruction.

 **/

namespace cxcparamx {

    void
    simple::_init_parfile( const char* cfilename,
			   int argc, char** argv, const char* mode ) {

	parfile = pParamFile( new ParamFile(cfilename, argc, argv, mode ) );
	filename =  paramgetpath( parfile->pf );

    }

    // this is required so that auto_ptr to forward declared ParamFile class
    // is correctly destroyed.
    simple::~simple() { }

    /** \brief Constructor with file name taken from argv[0]
     *
     * \param[in] argc the number of elements in \c argv
     * \param[in] argv typically, the command line arguments vector passed to \c main*
     * \param[in] mode mode with which to open the file
    */
    simple::simple( int argc,  char** argv, const char* mode ) {
	_init_parfile( NULL, argc, argv, mode );
    }

    /** \brief Constructor with explicit parameter file name
     *
     * \param[in] filename the parameter file name
     * \param[in] argc the number of elements in \c argv
     * \param argv typically, the command line arguments vector passed to \c main
     * \param[in] mode mode with which to open the file
     */
    simple::simple( const string& filename,
		    int argc, char** argv, const char* mode ) {
	_init_parfile( filename.c_str(), argc, argv, mode );
    }


    /** \brief Constructor with explicit parameter file name

     * \param[in] filename the parameter file name
     * \param[in] argc the number of elements in \c argv
     * \param argv typically, the command line arguments vector passed to \c main
     * \param[in] mode mode with which to open the file
     */
    simple::simple( const char* filename,
		    int argc, char** argv, const char* mode ) {
	_init_parfile( filename, argc, argv, mode );
    }

    static
    void access( paramfile pf, const char* name, const string& filename ) {

	if ( ! paccess( pf, name ) )
	    throw( Exception( filename + ": no such parameter: " + name ) );

    }

    /** \brief Retrieve a parameter value into a float
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const char* name, float& value ) {
	access( parfile->pf, name, filename );
	value = pgetf( parfile->pf, name );
    }

    /** \brief Retrieve a parameter value into a double
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const char* name, double& value ) {
	access( parfile->pf, name, filename );
	value = pgetd( parfile->pf, name );
    }


    /** \brief Retrieve a parameter value into a short
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const char* name, short& value ) {
	access( parfile->pf, name, filename );
	value = pgets( parfile->pf, name );
    }

    /** \brief Retrieve a parameter value into an int
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const char* name, int& value ) {
	access( parfile->pf, name, filename );
	value = pgeti( parfile->pf, name );
    }

    /** \brief Retrieve a parameter value into a long
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const char* name, long& value ) {
	access( parfile->pf, name, filename );

	// pgetl is in cxcparam/parameter.h, but isn't in the actual
	// library
	value = pgeti( parfile->pf, name );
    }


    /** \brief Retrieve a parameter value into an bool
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const char* name, bool& value ) {
	access( parfile->pf, name, filename );
	value = pgetb( parfile->pf, name );
    }

    /** \brief Retrieve a parameter value into an string
     *
     * \param[in] name parameter name
     * \param[out]  value returned value
     * \throws Exception if parameter is not found
     */
    void
    simple::get( const char* name, string& value ) {

	access( parfile->pf, name, filename );

	const int strlen = 1024;
	char buffer[strlen+1];

	pgetstr( parfile->pf, name, buffer, strlen );
	value = buffer;
    }


    //////////////////////////////////////////////

    /** \brief Return a parameter value as a float
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    float
    simple::getf( const char* name ) {
	access( parfile->pf, name, filename );
	return pgetf( parfile->pf, name );
    }

    /** \brief Return a parameter value as a double
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    double
    simple::getd( const char* name ) {
	access( parfile->pf, name, filename );
	return pgetd( parfile->pf, name );
    }


    /** \brief Return a parameter value as a short
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    short
    simple::gets( const char* name ) {
	access( parfile->pf, name, filename );
	return pgets( parfile->pf, name );
    }

    /** \brief Return a parameter value as an int
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    int
    simple::geti( const char* name ) {
	access( parfile->pf, name, filename );
	return pgeti( parfile->pf, name );
    }

    /** \brief Return a parameter value as a long
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    long
    simple::getl( const char* name ) {
	access( parfile->pf, name, filename );

	// pgetl is in cxcparam/parameter.h, but isn't in the actual
	// library
	return pgeti( parfile->pf, name );
    }


    /** \brief Return a parameter value as an bool
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    bool
    simple::getb( const char* name ) {
	access( parfile->pf, name, filename );
	return pgetb( parfile->pf, name );
    }

    /** \brief Return a parameter value as an string
     *
     * \param[in] name parameter name
     * \throws Exception if parameter is not found
     */
    string
    simple::getstr( const char* name ) {

	access( parfile->pf, name, filename );

	const int strlen = 1024;
	char buffer[strlen+1];

	pgetstr( parfile->pf, name, buffer, strlen );
	return string(buffer);
    }



}
