// --8<--8<--8<--8<--
//
// Copyright (C) 2012 Smithsonian Astrophysical Observatory
//
// This file is part of imageblur
//
// imageblur is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// -->8-->8-->8-->8--

#include <iterator>
#include <algorithm>
#include <sstream>
#include <limits>

#include <randlib_c/randlib.h>
#include <Exception/Exception.h>

#include "fits_exception.h"
#include "cdf.h"
#include "fitsio2.h"

using namespace std;

void
CDF::dump( long max ) {

    Pixels::iterator end = max > 0 ? pixels.begin() + max : pixels.end();

    ostream_iterator<Pixel>err ( std::cerr, "\n" );
    copy( pixels.begin(), end, err );

}

void
CDF::dump( Pixels::iterator begin, Pixels::iterator end ) {

    ostream_iterator<Pixel>err ( std::cerr, "\n" );
    copy( begin, end, err );

}

void
CDF::dump( Pixels::iterator begin, long end ) {

    ostream_iterator<Pixel>err ( std::cerr, "\n" );
    copy( begin, begin + end, err );

}

void
CDF::read_image( string filename ) {

    // assume file is in FITS format.

    fitsfile *fptr;
    int status = 0;

    if ( fits_open_image( &fptr, filename.c_str(), READONLY, &status ) )
	throw( fits_exception("", filename ) );

    // make sure this is an image we can handle

    int bitpix;
    int naxis;
    long int extent[2];

    if ( fits_get_img_param( fptr, 2, &bitpix, &naxis, extent, &status ) )
	throw( fits_exception("", filename) );

    if ( naxis != 2 )
	throw( Exception( string("incorrect dimensions in image; must be exactly 2")
			  + " in " + filename ) );

    pixels.resize( extent[0] * extent[1] );

    // figure out how many rows to read at once. make sure number of
    // bytes is > MINDIRECT (from fitsio2.h) so CFITSIO uses direct
    // i/o rather than through buffers

    long int nrows = MINDIRECT / (extent[0] * sizeof(Pixel) ) + 1;
    long int npixels = nrows * extent[0];

    // read image in chunks into vector
    double *buffer = new double[npixels];

    long int rows_to_read;
    long int row;
    vector<Pixel>::iterator pptr = pixels.begin();

    for ( row = rows_to_read = 0; row < extent[1] ; row += rows_to_read ) {

	long int fpixel[2] = { 1, row + 1 };

	rows_to_read = row + nrows > extent[1] ? extent[1] - row : nrows;
	long int npix = rows_to_read * extent[0];

	fits_read_pix( fptr, TDOUBLE, fpixel, npix, NULL, buffer, NULL, &status);

	double* buffptr = buffer;

	// row is a unary based counter
	for ( long int trow = row ; trow < row + rows_to_read ; trow++ ) {

	    for ( int tcol = 0; tcol < extent[0] ; tcol++, buffptr++, pptr++ ) {

		(*pptr).x = tcol;
		(*pptr).y = trow;
		(*pptr).p = *buffptr;

	    }

	}
    }

    if ( row != extent[0] ) {

	std::ostringstream err;
	err << "read " << row << "; expected " << extent[0];
	throw( Exception( err.str() ) );

    }

    // now encode xoffset and yoffset
    if ( xoffset == numeric_limits<double>::max() )
	xoffset = extent[0] / 2;

    if ( yoffset == numeric_limits<double>::max() )
	yoffset = extent[1] / 2;

    delete[] buffer;
}

void
CDF::find_pixel(
		const double prob,
		long int& x,
		long int& y
		 ) {

    vector<Pixel>::iterator pos = lower_bound( pixels.begin(), pixels.end(), prob );

    if ( pos == pixels.end() ) {
	std::ostringstream err;
	err << "internal error: unable to find probability " << prob << " in image";
	throw( Exception( err.str() ) );
    }

    x = pos->x;
    y = pos->y;
}

void
CDF::offset( double& x, double& y ) {

    long int ix, iy;

    find_pixel( ranf(), ix, iy );

    x += scale * ( ix - xoffset + ( 0.5 - ranf() ) );
    y += scale * ( iy - yoffset + ( 0.5 - ranf() )  );
}


CDF::CDF( string filename,
	  double xoffset, double yoffset, double scale,
	  double pmin, double pmax, string clip ) :
    xoffset(xoffset), yoffset(yoffset), scale(scale)
{

    read_image( filename );

    // if we're clipping in probability space, need to perform surgery on list of pixels
    if ( pmin > 0 || pmax < 1 ) {

	if ( clip == "pdf" ) {

	    // clip the PDF

	    // pmin & pmax are fractions; scale by integral to get values
	    // in image units.
	    double sum = for_each( pixels.begin(), pixels.end(), Pixel::sum() );

	    Pixels::iterator pos = remove_if( pixels.begin(), pixels.end(),
					      Pixel::clip( pmin * sum, pmax * sum) );
	    if ( pos != pixels.end() )
		pixels.erase( pos, pixels.end() );

	    // now, integrate the pdf
	    sort( pixels.begin(), pixels.end() );
	    for_each( pixels.begin(), pixels.end(), Pixel::partial_sum( ) );

	    // and normalize it
	    for_each( pixels.begin(), pixels.end(), Pixel::divides( pixels.back().p ) );

	} else if ( clip == "cdf" ) {

	    // clip the CDF

	    // integrate the pdf
	    sort( pixels.begin(), pixels.end() );
	    double sum = for_each( pixels.begin(), pixels.end(), Pixel::partial_sum( ) );

	    // pmin & pmax are fractions; scale by integral to get values
	    // in image units.

	    // upper_bound returns first element > value
	    Pixels::iterator min = upper_bound( pixels.begin(), pixels.end(), pmin * sum );
	    Pixels::iterator max = upper_bound( pixels.begin(), pixels.end(), pmax * sum );

	    // erase top part of CDF
	    if ( max != pixels.end() )
		pixels.erase( max, pixels.end() );

	    // erase bottom part of CDF, need to subtract off (*min).p from all pixels
	    if ( min != pixels.begin() ) {
		double low = (min - 1)->p;
		pixels.erase( pixels.begin(), min );

		for_each( pixels.begin(), pixels.end(), Pixel::subtracts( low ) );
	    }

	    // renormalize
	    for_each( pixels.begin(), pixels.end(), Pixel::divides( pixels.back().p ) );

	} else {

	    throw( Exception( string("internal error: unknown value for clip: ") + clip ) );

	}



    } else {

	    // integrate the pdf
	    sort( pixels.begin(), pixels.end() );
	    for_each( pixels.begin(), pixels.end(), Pixel::partial_sum( ) );

	    // and normalize it
	    for_each( pixels.begin(), pixels.end(), Pixel::divides( pixels.back().p ) );

    }


}
