// --8<--8<--8<--8<--
//
// Copyright (C) 2012 Smithsonian Astrophysical Observatory
//
// This file is part of imageblur
//
// imageblur is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// -->8-->8-->8-->8--

#include <iostream>
#include <algorithm>
#include <numeric>
#include <cmath>
#include <functional>

#include <randlib_c/randlib.h>
#include <fits_exception.h>
#include <cdf.h>

#include "images.h"

#include "gtest/gtest.h"

struct expected_counts {

    long nevents;

    expected_counts(long nevents ) : nevents(nevents) {};

    void operator() (double &r) {

	r *= nevents;
    }

};

template <class T1, class T2>
struct gof2 : public std::binary_function<T1, T2, T1> {

    T1 operator() (T1 expected, T2 derived ) const {

	// use Gehrels variance function
	double sigma =  1 + sqrt(expected + 0.75);

	double gof = ( expected - derived ) / sigma;

	return gof*gof;
    }
};

void my_sqrt( double& x ) { x = sqrt(x) ; }

void
compare_vectors( vector<double>& v1, vector<double>& pdf, vector<double>& ratio ) {


    // calculate expected number of events based on total cast in v1

    // first generate proper probabilities
    vector<double> exp_n = pdf;
    double maxcdf = accumulate( exp_n.begin(), exp_n.end(), 0 );
    normalize_image( exp_n, maxcdf );

    long nevents = accumulate( v1.begin(), v1.end(), 0 );
    for_each( exp_n.begin(), exp_n.end(), expected_counts( nevents ) );

    ratio = v1;

    transform( exp_n.begin(), exp_n.end(), ratio.begin(), ratio.begin(), gof2<double,double>() );

    double red_chi2 = accumulate( ratio.begin(), ratio.end(), 0 ) / (double) v1.size() ;

    // for_each( ratio.begin(), ratio.end(), &my_sqrt );

    // this magic number represents the value derived for the fiducial
    // run.
    ASSERT_DOUBLE_EQ( 0.5778, red_chi2 );

}



void
compare_images( string f1, string f2, string f3 ) {

    vector<double> image1 = read_image( f1 );
    vector<double> image2 = read_image( f2 );

    long extent[2];
    get_image_info( f1, extent );

    vector<double> ratio;
    compare_vectors( image1, image2, ratio );

    write_image( f3, ratio, extent );
}


TEST( blur, image ) {

    string filename = "sequence.fits";

    CDF cdf( filename );
    long extent[2];

    get_image_info( filename, extent );

    long npix = extent[0] * extent[1];
    vector<double> image(npix);

    fill( image.begin(), image.end(), 0 );

    for ( int i = 0; i < npix * 1000 ; i++ ) {

	long int x, y;
	cdf.find_pixel( ranf(), x, y );

	ASSERT_LT( x, extent[0]);
	ASSERT_LT( y, extent[1]);

	image[ y * extent[1] + x ] ++;
    }

    //    write_image( "raw.fits", image, extent );

    vector<double> ratio;

    compare_vectors( image, read_image( filename), ratio );

    //    write_image( "ratio.fits", ratio, extent );
}
