#include <tparam/tparam.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

/*************************************************************************/

static void gparam_ascii(paramfile pf);
static int setup_hdr(FILE *stream, PixStorage pix_storage,
		     long i_min, long i_max, long j_min, long j_max);
static int setup_raw(FILE *stream, PixStorage pix_storage,
		     long i_min, long i_max, long j_min, long j_max);
static void ascii_hdr_start(size_t len);
static void ascii_hdr_entry(char *var, char *format, ...);
static void ascii_hdr_end(void);
static void start_i(long i, size_t nj);
static void write_pix(CoreWeight *pix, long j_min, size_t npix);
static void write_zero(long j_min, size_t npix);
static void end_i(void);
static void cleanup(void);

/*************************************************************************/

ImageDump
dump_ascii_raw =
{
  "ascii-raw",		/* the unique identifier for an image dump */
  gparam_ascii,		/* parameter parser */
  setup_raw,		/* setup */
  ascii_hdr_start,
  ascii_hdr_entry,
  ascii_hdr_end,
  NULL,
  start_i,
  write_pix,
  write_zero,
  end_i,
  cleanup
};

ImageDump
dump_ascii_hdr =
{
  "ascii-hdr",		/* the unique identifier for an image dump */
  gparam_ascii,		/* parameter parser */
  setup_hdr,		/* setup */
  ascii_hdr_start,
  ascii_hdr_entry,
  ascii_hdr_end,
  NULL,
  start_i,
  write_pix,
  write_zero,
  end_i,
  cleanup
};

/*************************************************************************/

/* dump method specific parameters set by gparam */

static char i_prefix[NAME_LENGTH+1];
static char i_suffix[NAME_LENGTH+1];

static char wt_format[NAME_LENGTH+1];
static char j_pad[NAME_LENGTH+1];

/*************************************************************************/

/* file global storage for methods */

/* the dump stream, set by setup */
static FILE *fdump = NULL;

/* the current value of i. set by start_i() */
static long current_i;

/* the number of pixels for the current value of i. initialized by start_i() */
static long n_pixels;

/* the number of pixels written for the current value of i. initialized by
   start_i(), updated by write_pix() and write_zero() */
static long nw_pixels;

/*************************************************************************/

#define DUMPFUNC "ASCII dump"

static void
gparam_ascii(paramfile pf)
{
  tf_enter(DUMPFUNC " gparam");

  tpgetstr(pf, "i_prefix", i_prefix, NAME_LENGTH, 0);
  tpgetstr(pf, "i_suffix", i_suffix, NAME_LENGTH, 1);
  tpgetstr(pf, "wt_format", wt_format, NAME_LENGTH, 1);
  tpgetstr(pf, "j_pad", j_pad, NAME_LENGTH, 1);

  tf_leave();
}

static int
setup_hdr(
  FILE	     *stream,
  PixStorage  pix_storage,
  long	      i_min,
  long	      i_max,
  long	      j_min,
  long	      j_max
)
{
  tf_enter(DUMPFUNC " setup");

  fdump = stream;

  c.backend->write_header_ascii(pix_storage, i_min, i_max, j_min, j_max,
			      &dump_ascii_hdr);

  tf_leave();
  return 0;
}

static int
setup_raw(
  FILE	     *stream,
  PixStorage  pix_storage,
  long	      i_min,
  long	      i_max,
  long	      j_min,
  long	      j_max
)
{
  tf_enter(DUMPFUNC " setup");

  fdump = stream;

  tf_leave();
  return 0;
}

static void
ascii_hdr_start(size_t len)
{
  fprintf(fdump, "# %s\n", c.backend->name);
}

static void
ascii_hdr_entry(
  char *var,
  char *format,
  ...
)
{
  int ok;
  va_list ap;

  tf_enter(DUMPFUNC " ascii_hdr_entry");

  va_start(ap, format);

  ok =
    fprintf(fdump, "# %s = ", var) > 0 &&
    vfprintf(fdump, format, ap) > 0 &&
    fputc('\n', fdump) != EOF;

  va_end(ap);

  if (! ok)
    tf_exit(ExitERR_IOerr, "error writing header entry with format %s",
    				format);
  tf_leave();
}

static void
ascii_hdr_end(void)
{
}

static void
start_i(long i, size_t nj)
{
  tf_enter(DUMPFUNC " start_i");

  current_i = i;
  n_pixels = nj;
  nw_pixels = 0;

  if ( fputs(i_prefix, fdump) == EOF )
    tf_exit(ExitERR_fputs, "error at i=%d while writing i prefix", i);

  tf_leave();
}

static void
write_pix(CoreWeight *pix, long j_min, size_t npix)
{
  size_t n;

  tf_enter(DUMPFUNC " write_pix");

  for (n = 0 ; n < npix ; n++, pix++)
    if ( ( fprintf(fdump, wt_format, *pix) < 0 )
	||
	 ( ( ++nw_pixels < n_pixels ) && ( fputs(j_pad, fdump) == EOF ) )
	)
      tf_exit(ExitERR_fprintf,
         "error at (i=%d, j=%d) while writing pixel", current_i, j_min + n);

  tf_leave();
}

static void
write_zero(long j_min, size_t npix)
{
  size_t n;

  tf_enter(DUMPFUNC " write_zero");

  for (n = 0 ; n < npix ; n++)
  {
    if ( ( fprintf(fdump, wt_format, 0.0) < 0 )
	||
	 ( ( ++nw_pixels < n_pixels ) && ( fputs(j_pad, fdump) == EOF ) )
	)
      tf_exit(ExitERR_fprintf,
         "error at (i=%d, j=%d) while writing pixel", current_i, j_min + n);
  }

  tf_leave();
}

static void
end_i(void)
{
  tf_enter(DUMPFUNC " end_i");

  if ( fputs(i_suffix, fdump) == EOF )
    tf_exit(ExitERR_fputs, "error at i=%d while writing i suffix",
		    current_i);

  tf_leave();
}


static void
cleanup(void)
{
}
