#include <tparam/tparam.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

/*************************************************************************/

static void gparam_rdb(paramfile pf);
static int  setup_hdr(FILE *stream, PixStorage pix_storage, long i_min,
		      long i_max, long j_min, long j_max);
static int  setup_raw(FILE *stream, PixStorage pix_storage, long i_min,
		      long i_max, long j_min, long j_max);
static void setup_format(void);
static void rdb_hdr_start(size_t len);
static void rdb_hdr_entry(char *var, char *format, ...);
static void rdb_hdr_end(void);
static void rdb_hdr(void);
static void start_i(long i, size_t nj);
static void write_pix(CoreWeight *pix, long j_min, size_t npix);
static void write_zero(long j_min, size_t npix);
static void end_i(void);
static void cleanup(void);

/*************************************************************************/

ImageDump
dump_rdb_raw =
{
  "rdb-raw",		/* the unique identifier for an image dump */
  gparam_rdb,		/* parameter parser */
  setup_raw,		/* setup */
  rdb_hdr_start,
  rdb_hdr_entry,
  rdb_hdr_end,
  NULL,
  start_i,
  write_pix,
  write_zero,
  end_i,
  cleanup
};

ImageDump
dump_rdb_hdr =
{
  "rdb-hdr",		/* the unique identifier for an image dump */
  gparam_rdb,		/* parameter parser */
  setup_hdr,		/* setup */
  rdb_hdr_start,
  rdb_hdr_entry,
  rdb_hdr_end,
  NULL,
  start_i,
  write_pix,
  write_zero,
  end_i,
  cleanup
};

/*************************************************************************/

/* dump method specific parameters set by gparam */

static char wt_format[NAME_LENGTH+1];
static char pos_format[NAME_LENGTH+1];

/*************************************************************************/

/* file global storage for methods */

/* the dump stream, set by setup */
static FILE *fdump = NULL;

/* the current value of i. set by start_i() */
static long current_i;

/* the final format used to print out each pixel.  created by setup_format */
static char pix_format[2*NAME_LENGTH + 1];

/*************************************************************************/

#define DUMPFUNC "ASCII dump"

static void
gparam_rdb(paramfile pf)
{
  tf_enter(DUMPFUNC " gparam");

  tpgetstr(pf, "wt_format", wt_format, NAME_LENGTH, 1);
  tpgetstr(pf, "pos_format", pos_format, NAME_LENGTH, 1);

  tf_leave();
}

static int
setup_hdr(
  FILE	     *stream,
  PixStorage  pix_storage,
  long	      i_min,
  long	      i_max,
  long	      j_min,
  long	      j_max
)
{
  tf_enter(DUMPFUNC " setup");

  fdump = stream;

  c.backend->write_header_ascii(pix_storage, i_min, i_max, j_min, j_max,
			      &dump_rdb_hdr);

  setup_format();

  tf_leave();
  return 0;
}

static int
setup_raw(
  FILE	     *stream,
  PixStorage  pix_storage,
  long	      i_min,
  long	      i_max,
  long	      j_min,
  long	      j_max
)
{
  tf_enter(DUMPFUNC " setup");

  fdump = stream;

  setup_format();

  tf_leave();
  return 0;
}

static void
setup_format(void)
{
  sprintf(pix_format, "%%d\t%%d\t%s\t%s\t%s\n", pos_format, pos_format, wt_format);
}


static void
rdb_hdr_start(size_t len)
{
  fprintf(fdump, "# %s\n", c.backend->name);
}

static void
rdb_hdr_entry(
  char *var,
  char *format,
  ...
)
{
  int ok;
  va_list ap;

  tf_enter(DUMPFUNC " rdb_hdr_entry");

  va_start(ap, format);

  ok =
    fprintf(fdump, "#: %s = ", var) > 0 &&
    vfprintf(fdump, format, ap) > 0 &&
    fputc('\n', fdump) != EOF;

  va_end(ap);

  if (! ok)
    tf_exit(ExitERR_IOerr, "error writing header entry with format %s",
    				format);
  tf_leave();
}

static void
rdb_hdr_end(void)
{
}

static void
rdb_hdr(void)
{
  if ( fputs("i\tj\tx\ty\twt\n"
            "_\t_\t_\t_\t__\n", fdump) == EOF )
    tf_exit(ExitERR_IOerr, "error writing rdb header");
}

static void
start_i(long i, size_t nj)
{
  static int hdr_written = 0;
  tf_enter(DUMPFUNC " start_i");

  if (! hdr_written++)
    rdb_hdr();

  current_i = i;

  tf_leave();
}

static void
write_pix(CoreWeight *pix, long j_min, size_t npix)
{
  CoordIndex ci;
  long j;
  long j_lim = j_min + npix;

  tf_enter(DUMPFUNC " write_pix");

  for (j = j_min ; j < j_lim ; j++, pix++)
  {
    c.backend->bin_pos(current_i, j, &ci);
    if ( fprintf(fdump, pix_format, ci.i, ci.j, ci.x, ci.y, *pix) < 0 )
      tf_exit(ExitERR_fprintf,
         "error at (i=%d, j=%d) while writing pixel", ci.i, ci.j);
  }
  tf_leave();
}

static void
write_zero(long j_min, size_t npix)
{
  CoordIndex ci;
  long j;
  long j_lim = j_min + npix;

  tf_enter(DUMPFUNC " write_zero");

  for (j = j_min ; j < j_lim ; j++)
  {
    c.backend->bin_pos(current_i, j, &ci);
    if ( fprintf(fdump, pix_format, ci.i, ci.j, ci.x, ci.y, 0.0) < 0 )
      tf_exit(ExitERR_fprintf,
         "error at (i=%d, j=%d) while writing pixel", ci.i, ci.j);
  }

  tf_leave();
}

static void
end_i(void)
{
}


static void
cleanup(void)
{
}
