/*
   circular hexagonal grid back end routines.  this is a grid of circular
   bins on a grid such that adjacent rows are shifted in x by 0.5 grid
   spacings.  the diameter of the circles may be less than or equal to the
   grid spacing.
*/
#include <stddef.h>
#include <stdio.h>
#include <math.h>

#include <mathconst/mathconst.h>
#include <tparam/tparam.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

/*************************************************************************/

/* prototypes for statically defined functions in this file */

static void gparam_hexcirc(paramfile pf);
static int setup(void);
static int i_limits(long *min_i, long *max_i, unsigned long *nbins);
static void j_limits(long i, long *min_j, long *max_j);
static int clip(double x_in, double y_in, CoordIndex *ci);
static int noclip(double x_in, double y_in, CoordIndex *ci);
static int clip_packed(double x_in, double y_in, CoordIndex *ci);
static int noclip_packed(double x_in, double y_in, CoordIndex *ci);
static void bin_pos(long i, long j, CoordIndex *ci);
static double bin_area(long i, long j);
static void j_image_limits(long i, long j_min_e, long j_max_e, long *min_j,
			   long *max_j);
static void write_header_ascii(PixStorage pix_storage, long min_i, long max_i,
			       long min_j, long max_j, ImageDump *dump);
static void write_header_raw(PixStorage pix_storage, long min_i, long max_i,
			     long min_j, long max_j, ImageDump *dump);
static int cleanup(void);

/*************************************************************************/

BackEnd hexcirc_BackEnd =
{
  "hexcirc",			/* back end's name */
  BackEnd_HexCirc,		/* back end's id */
  gparam_hexcirc,		/* parameter read routine */
  setup,			/* setup routine */
  i_limits,			/* return primary index limits */
  j_limits,			/* return secondary index limits */
  noclip,			/* quantize routine (set by setup) */
  bin_pos,			/* return real coordinates from indices */
  bin_area,			/* return area of bin from indices */
  j_image_limits,		/* empirical secondary index limits */
  1,				/* front end should determine actual
  				   min and max secondary coordinate
  				   indices */
  write_header_raw,		/* write raw image header */
  write_header_ascii,		/* write ascii image header */
  cleanup			/* cleanup routine */
};

/*************************************************************************/

/*  file global storage for methods */

/* high and low i and j values */
static long i_max, i_min, j_max, j_min;

/* high and low x and y values */
static double y_max, y_min;
static double x_max_even, x_max_odd;
static double x_min_even, x_min_odd;

/* center of bin in input coordinates */
static double x_center_odd, x_center_even, y_center;

/* the rectangular "bin" dimensions */
static double x_sp, y_sp;
static double y_sp_h;

/* the region of the bin which is unique if the grid is packed */
static double unique, unique_top, unique_bot;

/*
   to determine the index of the pixel in which a value falls,
   you'd like to just do

              y_in / y_sp

   if the pixels are aligned such that the bottom edges fall upon
   multiples of the pixel size, then this works.  however, that
   doesn't always occur.  instead, we move the input coordinates
   into a coordinate system where pixel edges do lie on nice numbers.
*/
static double x_shift_odd, x_shift_even, y_shift;

/* the radius */
static double radius, radius2;

/* area of an aperture */
static double area;

/* the output header */
static HexCircImage hdr;
#define MapEntry(entry) {offsetof(HexCircImage, entry), sizeof(hdr.entry)}

static HdrMapEntry hdr_map[] =
{
  MapEntry(ni),
  MapEntry(nj),
  MapEntry(i_min),
  MapEntry(j_min),
  MapEntry(x0),
  MapEntry(y0),
  MapEntry(diameter),
  MapEntry(spacing),
  MapEntry(x_sp),
  MapEntry(y_sp)
};
#define n_entries (sizeof(hdr_map) / sizeof(HdrMapEntry))

#define SQRT3 1.73205080757

/*************************************************************************/

/* external parameters, filled in by gparam */
static int n_x, n_y;		/* the dimensions of the image */
static double diameter;		/* the diameter of the circles */
static double spacing;		/* the distance between centers */

/*************************************************************************/

static void
gparam_hexcirc(paramfile pf)
{
  tf_enter("gparam hexcirc");

  n_x = tpgetd(pf, "n_x");
  n_y = tpgetd(pf, "n_y");
  diameter = tpgetd(pf, "diameter");
  spacing = tpgetd(pf, "spacing");

  if (diameter > spacing)
    tf_exit(ExitERR_illegalArg,
      "inconsistent grid values:  diameter > spacing");

  tf_leave();
}

/*
     any initializations should be done here.
     return 0 on success, non zero on failure.
     note that x is indexed by j and y by i.
*/
static int
setup(void)
{

  radius = diameter / 2.0;
  radius2 = radius * radius;
  area = M_PI * radius2;

  x_sp = spacing;
  y_sp = SQRT3 * x_sp / 2;
  y_sp_h = y_sp / 2;
  unique = y_sp - radius;
  unique_top = y_sp_h + unique;
  unique_bot = y_sp_h - unique;

  if (c.clip)
  {
    double width, height;


    /* height and width of clipping region */
    width  = (n_x - 1) * x_sp + 2 * radius + ((n_y > 1) ? x_sp / 2 : 0);
    height = (n_y - 1) * y_sp + 2 * radius;

    x_center_even = width / 2 - ( ((n_x - 1) / 2) * x_sp + radius );
    x_center_odd = x_center_even - x_sp / 2;

    y_center = height / 2 - ( ((n_y - 1) / 2) * y_sp + radius);

    j_min = - (n_x / 2);
    j_max = j_min + n_x - 1;

    i_min = - (n_y / 2);
    i_max = i_min + n_y - 1;

    x_min_odd = j_min * x_sp + x_center_odd - radius + c.x_off;
    x_max_even = j_max * x_sp + x_center_even + radius + c.x_off;

    x_min_even = x_min_odd + x_sp / 2;
    x_max_odd = x_max_even - x_sp / 2;

    y_min = i_min * y_sp + y_center - radius + c.y_off;
    y_max = i_max * y_sp + y_center + radius + c.y_off;

    hexcirc_BackEnd.quantize = unique >= radius ? clip : clip_packed;
  }
  else
  {
    x_center_even = 0.0;
    x_center_odd = x_center_even - x_sp / 2;
    y_center = 0.0;
    hexcirc_BackEnd.quantize = unique >= radius ? noclip : noclip_packed;
  }

  x_shift_even = x_sp / 2 - x_center_even;
  x_shift_odd  = x_sp / 2 - x_center_odd;
  y_shift = y_sp_h - y_center;

  return 0;
}


/*
   returns the minimum and maximum bin index in the i dimension, as well
   as the total number of bins. returns 0 if all information is
   available, 1 if any is missing
*/
static int
i_limits(long *min_i, long *max_i, unsigned long *nbins)
{
  /* only know limits if clipping is in effect */
  if (c.clip)
  {
    *min_i = i_min;
    *max_i = i_max;
    *nbins = (i_max - i_min + 1) * (j_max - j_min + 1);
    return 0;
  }
  else
    return 1;
}


/*
   for a given i, return the minimum and maximum j values.  guaranteed
   to be called only with legal i values as determined by
   i_limits.  don't have to check for clipping here.
*/
static void
j_limits(long i, long *min_j, long *max_j)
{
  *min_j = j_min;
  *max_j = j_max;
}

/*
   returns the i, j pair corresponding to the passed position.
   returns the output position of the bin corresponding to the passed
   position.  returns 0 if the position is within bounds, non zero if
   not.
*/
static int
clip(double x_in, double y_in, CoordIndex *ci)
{
  double x, y, ys, xc, xs;
  double dx, dy;

  if (y_in < y_min || y_in > y_max)
    return 1;

  /* because the x min and max values depend upon whether it's an even
     or odd row, determine the row number first */

  ys = y_in - c.y_off + y_shift;

  ci->i = floor(ys / y_sp);
  if ( (y = ci->i * y_sp) > ys)
    y = --ci->i * y_sp;
  else if ((ci->i+1) * y_sp < ys)
    y = ++ci->i * y_sp;

  if (ci->i > i_max || ci->i < i_min)
    return 1;

  y += y_center + c.y_off;

  /* check if outside of circle */
  if (fabs(dy = y - y_in) > radius)
    return 1;

  if (abs(ci->i) % 2)
  {
    if (x_in < x_min_odd || x_in > x_max_odd) return 1;
    xs = x_in - c.x_off + x_shift_odd;
    xc = x_center_odd;
  }
  else
  {
    if (x_in < x_min_even || x_in > x_max_even) return 1;
    xs = x_in - c.x_off + x_shift_even;
    xc = x_center_even;
  }

  /* find index, worry about round off error */
  ci->j = floor(xs / x_sp);
  if ( (x = ci->j * x_sp) > xs)
    x = --ci->j * x_sp;
  else if ((ci->j+1) * x_sp < xs)
    x = ++ci->j * x_sp;

  x += xc + c.x_off;

  /* check if outside of circle */
  if (fabs(dx = x - x_in) > radius || dy * dy + dx * dx > radius2)
    return 1;

  ci->x = x;
  ci->y = y;

  return 0;
}

/*
   returns the i, j pair corresponding to the passed position.
   returns the output position of the bin corresponding to the passed
   position.  no clipping is done.
*/
static int
noclip(double x_in, double y_in, CoordIndex *ci)
{
  double x, y, xs, ys, xc;
  double dx, dy;

  /* because the x pixel edges depend upon whether it's an even or odd row,
     determine the row number first */

  ys = y_in - c.y_off + y_shift;

  ci->i = floor(ys / y_sp);
  if ( (y = ci->i * y_sp) > ys )
    y = --ci->i * y_sp;
  else if ( (ci->i+1) * y_sp < ys )
    y = ++ci->i * y_sp;

  y += y_center + c.y_off;

  /* check if outside of circle */
  if (fabs(dy = y - y_in) > radius)
    return 1;

  if (abs(ci->i) % 2)
  {
    xs = x_in - c.x_off + x_shift_odd;
    xc = x_center_odd;
  }
  else
  {
    xs = x_in - c.x_off + x_shift_even;
    xc = x_center_even;
  }

  /* find index, worry about round off error */
  ci->j = floor(xs / x_sp);
  if ( (x = ci->j * x_sp) > xs)
    x = --ci->j * x_sp;
  else if ((ci->j+1) * x_sp < xs)
    x = ++ci->j * x_sp;

  x += xc + c.x_off;

  /* check if outside of circle */
  if (fabs(dx = x - x_in) > radius || dy * dy + dx * dx > radius2)
    return 1;

  ci->x = x;
  ci->y = y;

  return 0;
}


#define IP(y)  floor((SQRT3 * y - xs) / spacing)
#define IPP(y) floor((SQRT3 * y + xs) / spacing)

static int
clip_packed(double x_in, double y_in, CoordIndex *ci)
{
  double x, y, ys, xc, xs;
  double dx, dy;

  if (y_in < y_min || y_in > y_max)
    return 1;

  ys = y_in - c.y_off + y_shift;	/* relative to edge of 0'th bin */

  ci->i = floor(ys / y_sp);
  if ( (y = ci->i * y_sp) > ys)
    y = --ci->i * y_sp;
  else if ((ci->i+1) * y_sp < ys)
    y = ++ci->i * y_sp;

  if ( ys > y + unique_top )
  {
    ys = y_in - c.y_off - y_center;
    xs = x_in - c.x_off - x_center_even;

    if ( IP(y) != IP(ys) && IPP(y) != IPP(ys))
      y = ++ci->i * y_sp;
  }
  else if ( ys < y + unique_bot )
  {
    ys = y_in - c.y_off - y_center;
    xs = x_in - c.x_off - x_center_even;

    if ( (IP(y) != IP(ys)) && (IPP(y) != IPP(ys)))
      y = --ci->i * y_sp;
  }

  if (ci->i > i_max || ci->i < i_min)
    return 1;

  y += y_center + c.y_off;

  /* check if outside of circle */
  if (fabs(dy = y - y_in) > radius)
    return 1;

  if (abs(ci->i) % 2)
  {
    if (x_in < x_min_odd || x_in > x_max_odd) return 1;
    xs = x_in - c.x_off + x_shift_odd;
    xc = x_center_odd;
  }
  else
  {
    if (x_in < x_min_even || x_in > x_max_even) return 1;
    xs = x_in - c.x_off + x_shift_even;
    xc = x_center_even;
  }

  /* find index, worry about round off error */
  ci->j = floor(xs / x_sp);
  if ( (x = ci->j * x_sp) > xs)
    x = --ci->j * x_sp;
  else if ((ci->j+1) * x_sp < xs)
    x = ++ci->j * x_sp;

  x += xc + c.x_off;

  /* check if outside of circle */
  if (fabs(dx = x - x_in) > radius || dy * dy + dx * dx > radius2)
    return 1;

  ci->x = x;
  ci->y = y;
  return 0;
}

static int
noclip_packed(double x_in, double y_in, CoordIndex *ci)
{
  double x, y, xs, ys, xc;
  double dx, dy;

  /* because the x pixel edges depend upon whether it's an even or odd row,
     determine the row number first */

  ys = y_in - c.y_off + y_shift;

  ci->i = floor(ys / y_sp);
  if ( (y = ci->i * y_sp) > ys)
    y = --ci->i * y_sp;
  else if ((ci->i+1) * y_sp < ys)
    y = ++ci->i * y_sp;

  if ( ys > y + unique_top )
  {
    ys = y_in - c.y_off - y_center;
    xs = x_in - c.x_off - x_center_even;

    if ( IP(y) != IP(ys) && IPP(y) != IPP(ys))
      y = ++ci->i * y_sp;
  }
  else if ( ys < y + unique_bot )
  {
    ys = y_in - c.y_off - y_center;
    xs = x_in - c.x_off - x_center_even;

    if ( IP(y) != IP(ys) && IPP(y) != IPP(ys))
      y = --ci->i * y_sp;
  }

  y += y_center + c.y_off;

  /* check if outside of circle */
  if (fabs(dy = y - y_in) > radius)
    return 1;

  if (abs(ci->i) % 2)
  {
    xs = x_in - c.x_off + x_shift_odd;
    xc = x_center_odd;
  }
  else
  {
    xs = x_in - c.x_off + x_shift_even;
    xc = x_center_even;
  }

  /* find index, worry about round off error */
  ci->j = floor(xs / x_sp);
  if ( (x = ci->j * x_sp) > xs)
    x = --ci->j * x_sp;
  else if ((ci->j+1) * x_sp < xs)
    x = ++ci->j * x_sp;

  x += xc + c.x_off;

  /* check if outside of circle */
  if (fabs(dx = x - x_in) > radius || dy * dy + dx * dx > radius2)
    return 1;

  ci->x = x;
  ci->y = y;

  return 0;
}

/*
     returns the output coordinates of the bin corresponding to
     the passed (i,j) pair
*/
static void
bin_pos(long i, long j, CoordIndex *ci)
{
  ci->i = i;
  ci->j = j;

  ci->y = ci->i * y_sp + y_center + c.y_off;
  if (abs(ci->j) % 2)
    ci->x = ci->j * x_sp + x_center_odd + c.x_off;
  else
    ci->x = ci->j * x_sp + x_center_even + c.x_off;
}

/*
     returns the area of the bin corresponding to the passed (i,j) pair
*/
static double
bin_area(long i, long j)
{
  return area;
}

/*
   given an i value and empiricially determined max and min j values,
   for an image, return j limits.  for rectangular grids, simply
   returning the passed limits should suffice.  for polar grids, the
   input limits have no real meaning.  when clipping is in effect,
   should return clipping limits.
*/
static void
j_image_limits(long i, long j_min_e, long j_max_e, long *min_j, long *max_j)
{
  if (c.clip)
  {
    *min_j = j_min;
    *max_j = j_max;
  }
  else
  {
    *min_j = j_min_e;
    *max_j = j_max_e;
  }
}

static void
fill_hdr(
  PixStorage pix_storage,
  long min_i,
  long max_i,
  long min_j,
  long max_j
)
{
  CoordIndex ci;

  tf_enter("hexcirc fill_header");

  bin_pos(0, 0, &ci);

  hdr.diameter = diameter;
  hdr.spacing = spacing;
  hdr.x_sp = x_sp;
  hdr.y_sp = y_sp;
  hdr.x0 = ci.x;
  hdr.y0 = ci.y;

  switch(pix_storage)
  {
    /* image is in core, must have been clipped */
    case CORE:
      min_i = i_min;
      max_i = i_max;
      min_j = j_min;
      max_j = j_max;
      break;

    case SPARSE:
      if (c.clip)
      {
	min_i = i_min;
	max_i = i_max;
	min_j = j_min;
	max_j = j_max;
      }
      break;

    default:
      tf_exit(ExitERR_misc, "internal error in %s at %d", __FILE__,
		      __LINE__);
  }

  hdr.i_min = min_i;
  hdr.ni    = max_i - min_i + 1;
  hdr.j_min = min_j;
  hdr.nj    = max_j - min_j + 1;

  tf_leave();
}

static void
write_header_ascii(
  PixStorage pix_storage,
  long min_i,
  long max_i,
  long min_j,
  long max_j,
  ImageDump *dump
)
{

  tf_enter("hexcirc write_header_ascii");

  fill_hdr(pix_storage, min_i, max_i, min_j, max_j);

  dump->ascii_hdr_start(0);

  dump->ascii_hdr_entry("ni",      "%d", hdr.ni);
  dump->ascii_hdr_entry("nj",      "%d", hdr.nj);
  dump->ascii_hdr_entry("i_min",   "%d", hdr.i_min);
  dump->ascii_hdr_entry("j_min",   "%d", hdr.j_min);
  dump->ascii_hdr_entry("x0",      "%f", hdr.x0);
  dump->ascii_hdr_entry("y0",      "%f", hdr.y0);
  dump->ascii_hdr_entry("diameter",  "%f", hdr.diameter);
  dump->ascii_hdr_entry("spacing", "%f", hdr.spacing);
  dump->ascii_hdr_entry("x_sp",    "%f", hdr.x_sp);
  dump->ascii_hdr_entry("y_sp",    "%f", hdr.y_sp);

  dump->ascii_hdr_end();

  tf_leave();
}


static void
write_header_raw(
  PixStorage  pix_storage,
  long	      min_i,
  long	      max_i,
  long	      min_j,
  long	      max_j,
  ImageDump  *dump
)
{
  tf_enter("hexcirc write_header_raw");

  fill_hdr(pix_storage, min_i, max_i, min_j, max_j);

  dump->binary_hdr(hexcirc_BackEnd.type, &hdr, sizeof(hdr), hdr_map,
		   n_entries);

  tf_leave();
}


/*
     any cleanups should be done here.
     return 0 on success, non zero on failure.
*/
static int
cleanup(void)
{
  return 0;
}
