/*
  These routines are used to generate a contigous image from a sparse
  matrix representation.  As these routines are called after all processing
  of the photons, the tree is deleted by these routiens.
*/

#include <stdlib.h>
#include <stdio.h>


#include <rbtree/rbtree.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"
#include "image_actions.h"

static void j_image_free(void *dp);


/*
   as the routines are called by the redblack routines with a very
   limited number of arguments, file-global variables are used to pass
   extra information to the action functions
*/

static FILE *image_stream;	 /* where the pixels are to go.  set by
				    image_free_setup */
static long next_j;		/* the next j value for the current i value */
static long next_i;		/* the next i value to process */

static long j_min, j_max, min_j, max_j, i_min, i_max;

/*
   set image stream for the tree dump routines, and record
   the j limits
*/
void
image_free_setup(long min_i, long max_i, long min_j, long max_j, FILE *stream)
{
  next_i = min_i;

  i_min = min_i;
  i_max = max_i;

  j_min = min_j;
  j_max = max_j;

  image_stream = stream;
}

/*
   dump out remainder of image with no filled pixels.  only really useful
   when clipping is enabled.
*/
void
image_free_cleanup(void)
{
  tf_enter("image_free_cleanup");

  for (; next_i <= i_max ; next_i++)
  {
    size_t n;

    c.backend->j_image_limits(next_i, j_min, j_max, &min_j, &max_j);

    n = max_j - min_j + 1;

    c.dump->start_i(next_i, n);
    c.dump->write_zero(min_j, n);
    c.dump->end_i();
  }

  tf_leave();
}

#ifdef DEBUG
static size_t nw_j;
#endif

/*
   this routine is called from the redblack delete routine.

   because the i-tree is sparse, this routine is responsible for writing
   out any missing rows between the current i-node and the previous
   i-node.  it relies upon the file-global variable next_i to tell
   it which i-node to process.
*/
void
i_image_free(void *dp)
{
  Inode *inode = (Inode *) dp;

#ifdef DEBUG
  size_t nw_i = 0;
#endif

  tf_enter("i_image_free");

  /* output rows between this i-node and the last */
  for (; next_i < inode->i ; next_i++)
  {
    size_t nw;

    c.backend->j_image_limits(next_i, j_min, j_max, &min_j, &max_j);

    nw = max_j - min_j + 1;

    c.dump->start_i(next_i, nw);
    c.dump->write_zero(min_j, nw);
    c.dump->end_i();

#ifdef DEBUG
    nw_i += nw;
#endif
  }

#ifdef DEBUG
  if (dbflag("nw"))
    tf_message("nw_i = %d\n", nw_i);
#endif

  c.backend->j_image_limits(next_i, j_min, j_max, &min_j, &max_j);
  next_j = min_j;

  c.dump->start_i(next_i, max_j - min_j + 1);

#ifdef DEBUG
  nw_j = 0;
#endif

  /* j_image_free takes care of any columns preceding the first j-node
     for this j-tree */
  rbtree_delete(inode->j, j_image_free, IN_ORDER, LEFT_TO_RIGHT);
  free(inode);


  /* this routine takes care of any columns after the last j-node, but
     still relevant to this i-tree */
  c.dump->write_zero(next_j, max_j - next_j + 1);
  c.dump->end_i();

#ifdef DEBUG
  nw_j += max_j - next_j + 1;
  if (dbflag("nw"))
    tf_message("nw_j = %d\n", nw_j);
#endif


  /* increment counter to indicate row after this one */
  next_i++;
  tf_leave();
}

/*
   dump out jnode, filling in for missing j values between
   last j_image_free call and current invocation
*/
static void
j_image_free(void *dp)
{
  size_t n;
  Jnode *jnode = (Jnode *) dp;

  n = jnode->j_min - next_j;

  c.dump->write_zero(next_j, n);

#ifdef DEBUG
  nw_j += n;
#endif

  n = jnode->j_max - jnode->j_min + 1;
  c.dump->write_pix(jnode->wt, jnode->j_min, n);

#ifdef DEBUG
  nw_j += n;
#endif

  next_j = jnode->j_max + 1;

  free(jnode->wt);
  free(jnode);
}
