#include <stdio.h>

#include <tparam/tparam.h>
#include <suplib/str.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

#include "image_write.h"

#define BUFFER_SIZE 8192

/*************************************************************************/
/* token lists for the parameter parsing */

/* the types of image datum allowed */
typedef enum
{
  ImageDatum_FLOAT,
  ImageDatum_DOUBLE
} ImageDatumType;

static TokListToken
image_datum[] =
{
  {"double",	ImageDatum_DOUBLE},
  {"float",	ImageDatum_FLOAT}
};
static TokList image_data = GenTokList(image_datum);

/* the language specific structure of binary image files */
typedef enum
{
  BinaryFile_C,
  BinaryFile_Fortran
} BinaryFileLanguage;

static TokListToken
file_language[] =
{
  {"c",		BinaryFile_C},
  {"fortran",	BinaryFile_Fortran}
};
static TokList file_languages = GenTokList(file_language);

/*************************************************************************/
/* parameters determined by gparam_image_datum */

static size_t sizeof_image_datum;	/* the size of an image datum */
static ImageDatumType datum_type;	/* the type of an image datum */
static BinaryFileLanguage file_lang;	/* the language of the binary file */

/*************************************************************************/
/* file global variables for methods */

/* buffer for conversion of core to image data */
static void *conversion_buffer = NULL;

/* buffer for efficient writes of empty pixels */
static void *zerobuffer = NULL;

/*************************************************************************/

static int   write_pix(FILE *image, CoreWeight *data, size_t n);
static int   write_pix_float(FILE *image, CoreWeight *data, size_t n);
static int   write_pix_double(FILE *stream, CoreWeight *data, size_t n);
static int   write_zero(FILE *stream, size_t n);
static void  F_hdr(FILE *stream, BackEndType type, void *hdr,
		   size_t sizeof_hdr, HdrMapEntry map[], size_t n_entries);
static int   F_start_i(FILE *stream, size_t n);
static int   F_end_i(FILE *stream);
static void  C_hdr(FILE *stream, BackEndType type, void *hdr,
		   size_t sizeof_hdr, HdrMapEntry map[], size_t n_entries);
static int   C_start_i(FILE *stream, size_t n);
static int   C_end_i(FILE *stream);

/*************************************************************************/

/*
   get parameters relevant to the low level structure of the image file
*/
void
gparam_image_write(paramfile pf)
{
  char dummy[NAME_LENGTH+1];

  tf_enter("gparam");

  tpgetstr(pf, "image_datum", dummy, NAME_LENGTH, 1);
  datum_type = tokmatch(dummy, &image_data);
  if (datum_type == -1)
    tf_exit(ExitERR_illegalArg,
		    "illegal image_datum type: `%s'", dummy);

  tpgetstr(pf, "image_lang", dummy, NAME_LENGTH, 1);
  file_lang = tokmatch(dummy, &file_languages);
  if (file_lang == -1)
      tf_exit(ExitERR_illegalArg,
		      "illegal image_datum type: `%s'", dummy);

  tf_leave();
}


/*************************************************************************/

/*
   set up the methods to write the correct binary format (C or Fortran)
   and to perform mappings between the core pixel data type and the
   image pixel data type
 */
void
image_write_setup(void)
{
  tf_enter("setup_image_datum");

  switch (datum_type)
  {
    case ImageDatum_DOUBLE:
      sizeof_image_datum = sizeof(double);
      break;

    case ImageDatum_FLOAT:
      sizeof_image_datum = sizeof(float);
      break;

    default:
      tf_exit(ExitERR_misc, "internal error in %s at %d", __FILE__,
		      __LINE__);
  }

  zerobuffer = malloc(BUFFER_SIZE * sizeof_image_datum);
  if (!zerobuffer)
    tf_exit(ExitERR_alloc, "couldn't allocate image write buffer");

  if (datum_type == ImageDatum_DOUBLE)
  {
    double *d = (double *) zerobuffer;
    double *dlim = d + BUFFER_SIZE;
    while (d < dlim)
      *d++ = 0.0;
  }
  else
  {
    float *f = (float *) zerobuffer;
    float *flim = f + BUFFER_SIZE;
    while (f < flim)
      *f++ = 0.0;
  }

  if (sizeof_image_datum == sizeof(CoreWeight))
  {
    c.binary.write_pix = write_pix;
  }
  else
  {
    conversion_buffer = malloc(BUFFER_SIZE * sizeof_image_datum);

    if (! conversion_buffer)
      tf_exit(ExitERR_alloc, "couldn't allocate image write buffer");

    if (datum_type == ImageDatum_DOUBLE)
      c.binary.write_pix = write_pix_double;
    else
      c.binary.write_pix = write_pix_float;

  }

  c.binary.write_zero = write_zero;

  switch (file_lang)
  {
    case BinaryFile_C:
      c.binary.hdr       = C_hdr;
      c.binary.start_i   = C_start_i;
      c.binary.end_i     = C_end_i;
      break;

    case BinaryFile_Fortran:
      c.binary.hdr       = F_hdr;
      c.binary.start_i   = F_start_i;
      c.binary.end_i     = F_end_i;
      break;
  }

  tf_leave();
}


/*
   cleanup after all of the image writing is done
*/
void
image_write_cleanup(void)
{
  if (conversion_buffer) free(conversion_buffer);
  if (zerobuffer) free(zerobuffer);
}

/*************************************************************************/

/*
   only one of the the following methods is used to write data out,
   depending upon whether the core pixel data type is the same as the
   image pixel data type
 */

/* the core pixel type is the same as the image pixel type */
static int
write_pix(FILE *image, CoreWeight *data, size_t n)
{
  return fwrite(data, sizeof(CoreWeight), n, image) != n;
}

/* the image pixel type is float */
static int
write_pix_float(FILE *image, CoreWeight *data, size_t n)
{
  size_t n_left = n;
  size_t nwrite = n > BUFFER_SIZE ? BUFFER_SIZE : n;

  while (n_left)
  {
    size_t nw;
    size_t nout = n_left > nwrite ? nwrite : n_left;

    float *p = (float *) conversion_buffer;
    float *plim = p + nout;
    while (p < plim) *p++ = *data++;

    n_left -= (nw = fwrite(conversion_buffer, sizeof(float), nout, image));
    if (nw != nout) break;
  }

  return n_left != 0;
}

/* the image pixel type is double */
static int
write_pix_double(FILE *stream, CoreWeight *data, size_t n)
{
  size_t n_left = n;
  size_t nwrite = n > BUFFER_SIZE ? BUFFER_SIZE : n;

  while (n_left)
  {
    size_t nw;
    size_t nout = n_left > nwrite ? nwrite : n_left;

    double *p = (double *) conversion_buffer;
    double *plim = p + nout;
    while (p < plim) *p++ = *data++;

    n_left -= (nw = fwrite(conversion_buffer, sizeof(double), nout, stream));
    if (nw != nout) break;
  }

  return n_left != 0;
}

/*************************************************************************/

/*
   efficient way to write out empty pixels
*/
static int
write_zero(FILE *stream, size_t n)
{
  size_t nwrite;
  size_t n_left = n;

  nwrite = n > BUFFER_SIZE ? BUFFER_SIZE : n;
  while (n_left)
  {
    size_t nw;
    size_t nout = n_left > nwrite ? nwrite : n_left;
    n_left -= (nw = fwrite(zerobuffer, sizeof_image_datum, nout,
			   stream));
    if (nw != nout) break;
  }

  return n_left != 0;
}


/*************************************************************************/
/*************************************************************************/

/*
   The following routines are used to write out C and Fortran versions of
   a binary record.  Image dump headers are treated differently from
   pixel records.  The correct set of routines are stuffed into the
   global structure c.binary by image_write_setup().
*/


/* the length of the current record, in bytes */
static size_t F_record_length;

static int
F_write_reclen(FILE *stream, size_t len)
{
  int length = (int) len;	/* should be INTEGER *4 */
  return fwrite(&length, sizeof(length), 1, stream) != 1;
}

static void
F_hdr(FILE *stream, BackEndType type, void *hdr, size_t sizeof_hdr,
      HdrMapEntry map[], size_t n_entries)
{
  size_t i;
  int typei = type;

  F_record_length = sizeof(int);

  if (F_write_reclen(stream, F_record_length))
    tf_exit(ExitERR_fwrite, "error writing start of header type record");
 
  if ( fwrite(&typei, sizeof(typei), 1, stream) != 1)
      tf_exit(ExitERR_fwrite, "error writing header type");

  if (F_write_reclen(stream, F_record_length))
    tf_exit(ExitERR_fwrite, "error writing end of header type record");

  F_record_length = 0;
  for (i = 0 ; i < n_entries ; i++)
    F_record_length += map[i].size;

  if (F_write_reclen(stream, F_record_length))
    tf_exit(ExitERR_fwrite, "error writing start of header record");
 
  for (i = 0; i < n_entries ; i++)
    if ( fwrite((char *) hdr + map[i].offset, map[i].size, 1, stream) != 1)
      tf_exit(ExitERR_fwrite, "error writing header entry %d", i+1);

  if (F_write_reclen(stream, F_record_length))
    tf_exit(ExitERR_fwrite, "error writing end of header record");
}

static int
F_start_i(FILE *stream, size_t n)
{
  F_record_length = n * sizeof_image_datum;
  return F_write_reclen(stream, F_record_length);
}

static int
F_end_i(FILE *stream)
{
  return F_write_reclen(stream, F_record_length);
}

static void
C_hdr(FILE *stream, BackEndType type, void *hdr, size_t sizeof_hdr,
      HdrMapEntry map[], size_t n_entries)
{
  int typei = type;
  
  tf_enter("C_hdr_entry");

  if ( fwrite(&typei, sizeof(typei), 1, stream) != 1)
      tf_exit(ExitERR_fwrite, "error writing header type");

  if ( fwrite(hdr, sizeof_hdr, 1, stream) != 1)
    tf_exit(ExitERR_fwrite, "error writing header");

  tf_leave();
}

static int
C_start_i(FILE *stream, size_t n)
{
  return 0;
}

static int
C_end_i(FILE *stream)
{
  return 0;
}
