#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <rbtree/rbtree.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"
#include "cmp.h"
#include "new_trees.h"

static void create_Jnode(RBTree tree, long j, CoreWeight wt);
static void join_Jnodes(RBTree tree, RBNode *n1, RBNode *n2);
static void extend_Jnode_min(RBTree tree, Jnode *jn, long j,
			     CoreWeight weight);
static void extend_Jnode_max(RBTree tree, Jnode *jn, long j,
			     CoreWeight weight);
#ifdef DEBUG
static CoreWeight sum_wt(Jnode *jn);
#endif

#if 0
#if defined(SunOS) && OSMAJREV < 5 && defined(__GNUC__)
void bcopy(char *b1, char *b2, int length);
#define memmove(dest,src,length) bcopy((char *) (src), (char *) (dest),(length))
#endif
#endif

RBTree
new_Itree(RBTree tree, long i)
{
  Inode *inode;

  tf_enter("create_row_tree");

  /* create new row entry in tree */
  if (! (inode = malloc(sizeof(Inode))) )
    tf_exit(ExitERR_alloc, "unable to allocate %u'th i",
                        rbtree_count(tree)+1);

  inode->i = i;

  if (rbtree_insert(tree, inode))
    tf_exit(ExitERR_misc, "unable to insert new i");

  if (! (inode->j = rbtree_new(Jcmp)) )
    tf_exit(ExitERR_alloc, "unable to allocate new j");

  tf_leave();
  return inode->j;
}


/* create new node, or append to existing, or join nodes */
void
new_Jnode(RBTree tree, RBNode *prev, RBNode *next, long j, double weight)
{
  size_t sizeof_RBNode = rbtree_node_size();
  Jnode *jnext;
  Jnode *jprev;

  tf_enter("new_Jnode");

  /* reduce memory costs for join or extending nodes.  it's preferred
     to join nodes that are close, so give 'em an extra boost
     (c.join_pad).  prefer extending to creating, so give those a
     boost (c.extend_pad) */

  /* neighbors exist on both sides */
  if (prev && next)
  {
    int min_cost;
    long cost[4];
    jnext = rbtree_node_get_data(next);
    jprev = rbtree_node_get_data(prev);

    cost[0] = (jnext->j_min - jprev->j_max - 1 - c.join_pad)
                   * sizeof(CoreWeight) - (sizeof(Jnode) + sizeof_RBNode);
    cost[1] = (jnext->j_min - j - c.extend_pad) * sizeof(CoreWeight);
    cost[2] = (j - jprev->j_max - c.extend_pad) * sizeof(CoreWeight);
    cost[3] = sizeof(CoreWeight) + sizeof(Jnode) + sizeof_RBNode;

    min_cost = 0;
    if (cost[1] < cost[min_cost]) min_cost = 1;
    if (cost[2] < cost[min_cost]) min_cost = 2;
    if (cost[3] < cost[min_cost]) min_cost = 3;

#ifdef DEBUG
    if (dbflag("costs"))
      fprintf(stderr, "costs: %ld %ld %ld %ld : %ld\n",
	      cost[0], cost[1], cost[2], cost[3], cost[min_cost]);
#endif

    switch(min_cost)
    {
      case 0:
        /* more efficient to join them */
      	join_Jnodes(tree, prev, next);
	jprev->wt[j - jprev->j_min] += weight;
	break;

      case 1:
	/* cheaper to extend the next node */
	extend_Jnode_min(tree, jnext, j, weight);
	break;

      case 2:
	/* cheaper to extend the previous node? */
	extend_Jnode_max(tree, jprev, j, weight);
	break;

      case 3:
	create_Jnode(tree, j, weight);
	break;
    }
  }
  /* only a previous neighbor */
  else if (prev)
  {
    jprev = rbtree_node_get_data(prev);

    /* cheaper to extend the previous node? */
    if ( (j - jprev->j_max - c.extend_pad) * (long) sizeof(CoreWeight) <=
	  (long) (sizeof(Jnode) + sizeof_RBNode + sizeof(CoreWeight)))
      extend_Jnode_max(tree, jprev, j, weight);

    /* create a new node */
    else
      create_Jnode(tree, j, weight);

  }
  /* only a next neighbor */
  else if (next)
  {
    jnext = rbtree_node_get_data(next);

    /* cheaper to extend the next node? */
    if ( (jnext->j_min - j - c.extend_pad) * (long) sizeof(CoreWeight) <=
	  (long) (sizeof(Jnode) + sizeof_RBNode + sizeof(CoreWeight)))
      extend_Jnode_min(tree, jnext, j, weight);

    /* create a new node */
    else
      create_Jnode(tree, j, weight);
  }
  /* no neighbors, create a new node */
  else
    create_Jnode(tree, j, weight);

  tf_leave();
}

static void
create_Jnode(RBTree tree, long j, CoreWeight weight)
{
  Jnode *jn;

  tf_enter("create_Jnode");

  if (! (jn = (Jnode *) malloc(sizeof(Jnode))) )
    tf_exit(ExitERR_alloc, "unable to allocate Jnode");

  jn->j_min = jn->j_max = j;
  if (! (jn->wt = (CoreWeight *) malloc(sizeof(CoreWeight))) )
    tf_exit(ExitERR_alloc, "unable to allocate weight list");

  jn->wt[0] = weight;

  if (rbtree_insert(tree, jn))
    tf_exit(ExitERR_misc, "error inserting node into tree");

  tf_leave();
}

/*
    join adjacent nodes, creating space for intervening weights if necessary
*/
static void
join_Jnodes(RBTree tree, RBNode *n1, RBNode *n2)
{
  Jnode *j1, *j2;
  CoreWeight *wt;

  tf_enter("Join_Jnodes");

  j1 = rbtree_node_get_data(n1);
  j2 = rbtree_node_get_data(n2);

  wt = (CoreWeight *) realloc(j1->wt, (j2->j_max - j1->j_min + 1) *
			                         sizeof(CoreWeight));

  if (! wt)
    tf_exit(ExitERR_alloc, "unable to reallocate weights");

  j1->wt = wt;

  /* zero out any weights between the nodes */
  wt += j1->j_max - j1->j_min + 1;
  while (wt < j1->wt + (j2->j_min - j1->j_min)) *wt++ = 0.0;

  j1->j_max = j2->j_max;

  /* copy over weights from right node */
  memcpy(j1->wt + (j2->j_min - j1->j_min), j2->wt,
            (j2->j_max - j2->j_min + 1) * sizeof(CoreWeight));

  /* delete now useless node */
  rbtree_destroy_node(tree, n2);
  free(j2->wt);
  free(j2);

#ifdef DEBUG
  {
    CoreWeight *wt_lim = j1->wt + (j1->j_max - j1->j_min + 1);

    for ( wt = j1->wt ; wt < wt_lim ; wt++ )
      if ( *wt < 0.0 )
	abort( );
  }
#endif
  
  tf_leave();
}

/*
    extend a Jnode to a new minimum
*/
static void
extend_Jnode_min(RBTree tree, Jnode *jn, long j, CoreWeight weight)
{
  CoreWeight *wt;

  tf_enter("extend_Jnode_min");

  /* use realloc as it may use less memory than a calloc */

  wt = (CoreWeight *) realloc(jn->wt, (jn->j_max - j + 1) * sizeof(CoreWeight));

  if (!wt)
    tf_exit(ExitERR_alloc, "unable to reallocate weights");

  memmove(wt + (jn->j_min - j), wt, (jn->j_max - jn->j_min + 1) * sizeof(CoreWeight));
  jn->wt = wt;

  /* zero out new memory */
  while (wt < jn->wt + (jn->j_min - j)) *wt++ = 0.0;
  jn->j_min = j;

#ifdef DEBUG
  {
    CoreWeight *wt_lim = jn->wt + (jn->j_max - jn->j_min + 1);

    for ( wt = jn->wt ; wt < wt_lim ; wt++ )
      if ( *wt < 0.0 )
	abort( );
  }
#endif

  jn->wt[0] = weight;
  tf_leave();
}

/*
    extend a Jnode to a new maximum
*/
static void
extend_Jnode_max(RBTree tree, Jnode *jn, long j, CoreWeight weight)
{
  CoreWeight *wt;
#ifdef DEBUG
  CoreWeight old_wt;
#endif
  tf_enter("extend_Jnode_max");

#ifdef DEBUG
  if (dbflag("node_wt"))
    old_wt = sum_wt(jn);
#endif
  wt = (CoreWeight *) realloc(jn->wt,
			      (j - jn->j_min + 1) * sizeof(CoreWeight));
  if (!wt)
    tf_exit(ExitERR_alloc, "unable to reallocate weights");

  jn->wt = wt;

  /* zero out new memory */
  wt += jn->j_max - jn->j_min + 1;
  while (wt < jn->wt + (j - jn->j_min + 1)) *wt++ = 0.0;
  jn->j_max = j;

  jn->wt[j - jn->j_min] = weight;
#ifdef DEBUG
  if (dbflag("node_wt") && old_wt + weight != sum_wt(jn))
    die("old_wt + weight != sum_wt");
  {
    CoreWeight *wt_lim = jn->wt + (jn->j_max - jn->j_min + 1);

    for ( wt = jn->wt ; wt < wt_lim ; wt++ )
      if ( *wt < 0.0 )
	abort( );
  }
#endif
  tf_leave();
}

#ifdef DEBUG
static CoreWeight
sum_wt(Jnode *jn)
{
  long j;
  CoreWeight weight = 0.0;
  for (j = jn->j_min ; j <= jn->j_max ; j++)
    weight += jn->wt[j - jn->j_min];
  return weight;
}
#endif
