#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <limits.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <rbtree/rbtree.h>             /* binary tree stuff */
#include <tracefct/tracefct.h>           /* function stack stuff */
#include <tracefct/exiterrvals.h>        /* more function stack stuff */
#include <suplib/debug.h>
#include <suplib/times.h>

#include "main.h"
#include "common.h"

#include "process.h"
#include "core_image_dump.h"
#include "sparse_image_dump.h"

#include "cmp.h"
#include "rwbpipe.h"
#include "stream_actions.h"
#include "norm_actions.h"
#include "new_trees.h"
#include "debug.h"

typedef struct
{
  CoreWeight *pix;
  long i;
  long min_j;
  long max_j;
} CoreI;

static void normalize(PixStorage, CoreI *, RBTree,  size_t);

void
process(void)
{
  CoreWeight *fpix = NULL;	/* in core pixel array */
  unsigned long nphot = 0;
  unsigned long npix, n_i = 0;
  PixStorage pix_storage;
  RBTree Itree;

  FILE *fimage = NULL;		/* stream handle for image file */
  int create_image = 0;
  CoreI *core_i = NULL;

  int norm_bins = 0;

  /*
     minimum and maximum values of first grid coordinate, as returned by
     binning routine i_limits
     */
  long min_i, max_i;

  Bpipe bpipe;
  BpipePhoton photon;

  tf_enter("process");

  init_bpipe( c.input, c.output, &bpipe, &photon );

  /* make sure that image output stream is valid */
  if ( c.dump_image &&
      ! (fimage = strcmp(c.image_file, "stdout") ?
	     (strcmp(c.image_file, "stderr") ? fopen(c.image_file, "w") : stderr)
	      : stdout)
     )
    tf_exit(ExitERR_fopen, "couldn't open image `%s'", c.image_file);

  /* allow binning back end to perform initialization */
  c.backend->setup();

  /*
    if creating image or in filter_bin mode, the photons have to be
    accumulated.  if clipping is specified and the size of the grid
    is provided, it's possible to figure out if in-core storage
    is possible.  if there isn't enough memory, or there's not
    enough info, the "sparse marix" tree method is used.
   */

  if (c.dump_image || c.filter == FilterMode_BIN)
  {
    create_image++;

    if (c.clip &&
         ! c.backend->i_limits(&min_i, &max_i, &npix) &&
       ( (max_i - min_i + 1) * sizeof (CoreI) + npix * sizeof(CoreWeight) )
              < c.bufsz)
      pix_storage = CORE;
    else
      pix_storage = SPARSE;

    /* set up storage area, either tree or core */
    switch(pix_storage)
    {
      case SPARSE:
        /* create i tree */
        if (! (Itree = rbtree_new(Icmp)) )
          tf_exit(ExitERR_alloc, "couldn't create i tree");
        break;

      case CORE:
        n_i = max_i - min_i + 1;

        /* allocate block */
        if (! (fpix = (CoreWeight *) malloc(npix * sizeof(CoreWeight))))
          tf_exit(ExitERR_alloc, "unable to allocate image buffer");

	/* zero block */
        {
	  CoreWeight *pix = fpix;
	  CoreWeight *pixlim = fpix + npix;
	  while (pix < pixlim) *pix++ = 0.0;
	}
        /* allocate i pointers */
        if (! (core_i = (CoreI *) malloc(n_i * sizeof(CoreI))) )
          tf_exit(ExitERR_alloc,
			  "unable to allocate image buffer ptrs");

        /* set up CoreI information */
        {
	  long i;
          CoreI *iptr = core_i;
          CoreWeight *pix = fpix;

          for (i = min_i ; i <= max_i ; i++, iptr++)
          {
            iptr->i = i;
            c.backend->j_limits(i, &iptr->min_j, &iptr->max_j);
            iptr->pix = pix;
            pix += iptr->max_j - iptr->min_j + 1;
          }
        }
        break;

      default:
	tf_exit(ExitERR_misc, "internal error in %s at %d", __FILE__,
			__LINE__);
    }
  }

  if (dbflag("time"))
    startTime(NULL);

  while( bpipe_read_dpkts( bpipe.bpipe, bpipe.data, 1 ) ) {

    CoordIndex ci;
    double x, y, z, wt;
    Inode *inode;
    RBNode rbnode, prev, next;
    RBTree Jtree;
    int blocked;

#ifdef DEBUG
    CoreWeight old_wt;
#endif

    nphot++;

    x  = *photon.x;
    y  = *photon.y;
    z  = *photon.z;
    wt = *photon.weight;

    blocked = c.backend->quantize(x, y, &ci);

    switch(c.filter)
    {
      case FilterMode_QUANTIZEPOS:
	if (! blocked)
	{
	  *photon.x = ci.x;
	  *photon.y = ci.y;
	  if ( bpipe_write_dpkt( bpipe.bpipe, bpipe.data, bpipe.outstream ) )
	    tf_exit(ExitERR_fwrite, "error writing photon %u to `%s'",
			    nphot, c.output);
	}
	break;
	
      case FilterMode_QUANTIZEIDX:
	if (! blocked)
	{
	  *photon.x = ci.j;
	  *photon.y = ci.i;
	  if ( bpipe_write_dpkt( bpipe.bpipe, bpipe.data, bpipe.outstream ) )
	    tf_exit(ExitERR_fwrite, "error writing photon %u to `%s'",
			    nphot, c.output);
	}
	break;
	
      case FilterMode_PASS:
	if ( bpipe_write_dpkt( bpipe.bpipe, bpipe.data, bpipe.outstream ) )
	  tf_exit(ExitERR_fwrite, "error writing photon %u to `%s'",
			  nphot, c.output);
	break;

      case FilterMode_MASK:
	if (!blocked && 
	    bpipe_write_dpkt( bpipe.bpipe, bpipe.data, bpipe.outstream ) )
	  tf_exit(ExitERR_fwrite, "error writing photon %u to `%s'",
		  nphot, c.output);
	break;

      case FilterMode_NEGMASK:
	if (blocked &&
	    bpipe_write_dpkt( bpipe.bpipe, bpipe.data, bpipe.outstream ) )
	  tf_exit(ExitERR_fwrite, "error writing photon %u to `%s'",
		  nphot, c.output);
	break;

      case FilterMode_NONE:
	break;

      case FilterMode_BIN:
	break;
	
      default:
	tf_assert( 0 );
    }

    if (blocked) continue;

    if (dbflag("nphots"))
      tf_message("nphot: %ld i: %ld j: %ld\n", nphot, ci.i, ci.j);


    if (create_image)
      switch(pix_storage)
      {
        case CORE:
	  core_i[ci.i-min_i].pix[ci.j-core_i[ci.i-min_i].min_j] += wt;
	  break;

	case SPARSE:

#ifdef DEBUG
	  if (dbflag("tree_wt"))
	    old_wt = tree_wt(Itree);
#endif
	  /* search i tree */
	  if ( (inode = rbtree_search(Itree, &ci.i, Icmp_l)) )
	    Jtree = inode->j;
	  else
	    if (! (Jtree = new_Itree(Itree, ci.i)) )
	      tf_exit(ExitERR_misc, "error return from new_Itree");

	  if ((rbnode =
		     rbtree_bnd_search_node(Jtree, &ci.j, &prev, &next, Jcmp_l)))
	  {
	    Jnode *jnode = rbtree_node_get_data(rbnode);
	    jnode->wt[ci.j-jnode->j_min] += wt;
	  }
	  else
	  {
	    new_Jnode(Jtree, prev, next, ci.j, wt);
#ifdef DEBUG
	    if (dbflag("new_dump"))
	      tree_dump(Itree);
#endif
	  }
#ifdef DEBUG
	  if (dbflag("tree_wt") && fabs(old_wt + wt - tree_wt(Itree)) > 0.001)
	  {
	    fprintf(stderr, "weights out of line at nphot=%ld\n", nphot);
	    fprintf(stderr, "old_wt: %f wt: %f  = %f new_wt: %f\n", old_wt, wt,
		    old_wt + wt, tree_wt(Itree));
	    exit(1);
	  }
#endif
	  break;

	default:
	  tf_exit(ExitERR_misc, "internal error in %s at %d", __FILE__,
			  __LINE__);
      }
  }

  /*
  if (ferror(fin))
    tf_exit(ExitERR_fread, "error in input file `%s' at ray %u",
               c.input, nphot+1);
	       */

  if (dbflag("time"))
    tf_message("time: %s\n", stringTime(elapsedTime(NULL)));

  if (dbflag("tree_dump") && create_image && pix_storage == SPARSE)
    tree_dump(Itree);

  if (dbflag("efficiency") && create_image && pix_storage == SPARSE)
    tree_efficiency(Itree);


  if (!norm_bins && (c.norm & Normalize_IMAGE && c.norm & Normalize_BIN) )
  {
    normalize(pix_storage, core_i, Itree, n_i);
    norm_bins++;
  }

  /*
     if in filter_bin mode, have accumulated photons.
     here's where it gets dumped */
  if (c.filter == FilterMode_BIN)
  {

    switch(pix_storage)
    {
      case CORE:
        /* count number of binned photons */
        {
          CoreWeight *pix = fpix;
          CoreWeight *pix_lim = fpix + npix;
        }

        {
          CoreI *iptr = core_i;
          long i;

          for (i = 0 ; i < n_i ; i++, iptr++)
          {
	    CoordIndex ci;
            CoreWeight *pix = iptr->pix;
	    long j;

	    if (! norm_bins && (c.norm & Normalize_BIN) )
	    {
	      for (j = iptr->min_j ; j <= iptr->max_j ; j++, pix++)
		if (*pix)
		{
		  c.backend->bin_pos(iptr->i, j, &ci);
		  *photon.x = ci.x;
		  *photon.y = ci.y;
		  *photon.z = 0.0;
		  *photon.weight = *pix / c.backend->bin_area(iptr->i, j);
		  if ( bpipe_write_dpkt( bpipe.bpipe, bpipe.data,
					 bpipe.outstream ) )
		    tf_exit(ExitERR_fwrite, "error writing photon %u to `%s'",
			    nphot, c.output);
		}
	    }
	    else
	    {
	      for (j = iptr->min_j ; j <= iptr->max_j ; j++, pix++)
		if (*pix)
		{
		  c.backend->bin_pos(iptr->i, j, &ci);
		  *photon.x = ci.x;
		  *photon.y = ci.y;
		  *photon.z = 0.0;
		  *photon.weight = *pix;
		  if ( bpipe_write_dpkt( bpipe.bpipe, bpipe.data,
					 bpipe.outstream ) )
		    tf_exit(ExitERR_fwrite, "error writing photon %u to `%s'",
			    nphot, c.output);
		}
	    }
          }
        }
        break;

      case SPARSE:
        /*
	   traverse tree, dumping out data.  if we don't need to
	   create an image, delete the tree along the way.  first tell
	   tree dump routines where to dump
	 */
        set_dump_stream(&bpipe, &photon, 
			! norm_bins && (c.norm & Normalize_BIN));
        if (! c.dump_image)
            rbtree_delete(Itree, i_pos_free, POST_ORDER, LEFT_TO_RIGHT);
        else
            rbtree_traverse(Itree, i_pos_out, LEFT_TO_RIGHT);
        break;

      default:
	tf_exit(ExitERR_misc, "internal error in %s at %d",
			__FILE__, __LINE__);
    }

  }

  bpipe_delete( bpipe.bpipe );

  if (c.dump_image)
  {
    if (!norm_bins && (c.norm & Normalize_IMAGE) )
    {
      normalize(pix_storage, core_i, Itree, n_i);
      norm_bins++;
    }

    switch(pix_storage)
    {
      case CORE:
        core_image_dump(fpix, npix, min_i, max_i, fimage);
        break;

      case SPARSE:
	sparse_image_dump(Itree, min_i, max_i, fimage);
        break;

      default:
	tf_exit(ExitERR_misc, "internal error in %s at %d", __FILE__,
			__LINE__);
    }
    fclose(fimage);
  }

  if (fpix) free(fpix);
  if (core_i) free(core_i);
  c.backend->cleanup();

  tf_leave();

}


  /*
     here's where the accumulated photons get normalized by the area
     of the bins, if so requested.  it'd be more efficient to combine
     the process with one of the dump routines, but it'd be much messier to
     implement
   */
static void
normalize(PixStorage pix_storage, CoreI *iptr, RBTree Itree, size_t n_i)
{
  switch(pix_storage)
  {
    case CORE:
    {
      long i;

      for (i = 0 ; i < n_i ; i++, iptr++)
      {
	CoreWeight *pix = iptr->pix;
	long j;

	for (j = iptr->min_j ; j <= iptr->max_j ; j++, pix++)
	  if (*pix)
	    *pix /= c.backend->bin_area(iptr->i, j);
      }
    }
    break;

    case SPARSE:
    rbtree_traverse(Itree, i_norm, LEFT_TO_RIGHT);

    default:
    tf_exit(ExitERR_misc, "internal error in %s at %d",
		    __FILE__, __LINE__);
  }
}
