/*
   rectcirc back end routines.
*/

#include <stdio.h>
#include <math.h>

#include <mathconst/mathconst.h>
#include <tparam/tparam.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

/*************************************************************************/

/* prototypes for statically defined functions in this file */

static void gparam_rectcirc(paramfile pf);
static int setup(void);
static int i_limits(long *min_i, long *max_i, unsigned long *nbins);
static void j_limits(long i, long *min_j, long *max_j);
static int clip(double x_in, double y_in, CoordIndex *ci);
static int noclip(double x_in, double y_in, CoordIndex *ci);
static void bin_pos(long i, long j, CoordIndex *ci);
static double bin_area(long i, long j);
static void j_image_limits(long i, long j_min_e, long j_max_e, long *min_j,
			   long *max_j);
static void write_header_ascii(PixStorage pix_storage, long min_i, long max_i,
			       long min_j, long max_j, ImageDump *dump);
static void write_header_raw(PixStorage pix_storage, long min_i, long max_i,
			     long min_j, long max_j, ImageDump *dump);
static int cleanup(void);



/*************************************************************************/

BackEnd rectcirc_BackEnd =
{
  "rectcirc",			/* back end's name */
  BackEnd_RectCirc,		/* back end's id */
  gparam_rectcirc,		/* parameter read read routine */
  setup,			/* setup routine */
  i_limits,			/* return primary index limits */
  j_limits,			/* return secondary index limits */
  noclip,			/* quantize routine (set by setup) */
  bin_pos,			/* return real coordinates from indices */
  bin_area,			/* return area of bin from indices */
  j_image_limits,		/* empirical secondary index limits */
  1,				/* front end should determine actual
  				   min and max secondary coordinate
  				   indices */
  write_header_raw,		/* write raw image header */
  write_header_ascii,		/* write ascii image header */
  cleanup			/* cleanup routine */
};

/*************************************************************************/
/*  file global storage for methods */

/* high and low i and j values */
static long i_max, i_min, j_max, j_min;

/* high and low x and y values */
static double x_max, x_min, y_max, y_min;

/* center of bin in input coordinates */
static double x_center, y_center;

static double radius, radius2;
static double area;

/* the amount to shift an incoming value so that simple division gives a
   correct index.  see noclip() for details */
static double x_shift, y_shift;

/* the output header */
static RectCircImage hdr;
#define MapEntry(entry) {offsetof(RectCircImage, entry), sizeof(hdr.entry)}

static HdrMapEntry hdr_map[] =
{
  MapEntry(ni),
  MapEntry(nj),
  MapEntry(i_min),
  MapEntry(j_min),
  MapEntry(x0),
  MapEntry(y0),
  MapEntry(diameter),
  MapEntry(x_sp),
  MapEntry(y_sp)
};
#define n_entries (sizeof(hdr_map) / sizeof(HdrMapEntry))

/*************************************************************************/

/* external parameters, filled in by gparam */
static int n_x, n_y;		/* the dimensions of the image */
static double diameter;		/* the diameter of the circles */
static double x_sp, y_sp;	/* the grid spacing */

/*************************************************************************/

static void
gparam_rectcirc(paramfile pf)
{
  tf_enter("gparam rectcirc");

  x_sp = tpgetd(pf, "x_sp");
  y_sp = tpgetd(pf, "y_sp");

  diameter = tpgetd(pf, "diameter");
  if (x_sp < diameter || y_sp < diameter)
    tf_exit(ExitERR_illegalArg,
	    "inconsistent grid; spacing must be greater than diameter");

  n_x = tpgeti(pf, "n_x");
  n_y = tpgeti(pf, "n_y");

  tf_leave();
}

/*
     any initializations should be done here.
     return 0 on success, non zero on failure.
     note that x is indexed by j and y by i.
*/
static int
setup(void)
{
  radius = diameter / 2.0;
  radius2 = radius * radius;
  area = M_PI * radius2;

  if (c.clip)
  {
    x_center = (n_x%2) ? 0.0 : x_sp / 2;
    y_center = (n_y%2) ? 0.0 : y_sp / 2;

    j_min = - (n_x / 2);
    j_max = j_min + n_x - 1;

    i_min = - (n_y / 2);
    i_max = i_min + n_y - 1;

    x_min = j_min * x_sp + x_center - radius + c.x_off;
    x_max = j_max * x_sp + x_center + radius + c.x_off;

    y_min = i_min * y_sp + y_center - radius + c.y_off;
    y_max = i_max * y_sp + y_center + radius + c.y_off;

    rectcirc_BackEnd.quantize = clip;
  }
  else
  {
    x_center = 0.0;
    y_center = 0.0;
    rectcirc_BackEnd.quantize = noclip;
  }

  x_shift = x_sp / 2 - x_center;
  y_shift = y_sp / 2 - y_center;

  return 0;
}


/*
   returns the minimum and maximum bin index in the i dimension, as well
   as the total number of bins. returns 0 if all information is
   available, 1 if any is missing
*/
static int
i_limits(long *min_i, long *max_i, unsigned long *nbins)
{
  /* only know limits if clipping is in effect */
  if (c.clip)
  {
    *min_i = i_min;
    *max_i = i_max;
    *nbins = (i_max - i_min + 1) * (j_max - j_min + 1);
    return 0;
  }
  else
    return 1;
}


/*
   for a given i, return the minimum and maximum j values.  guaranteed
   to be called only with legal i values as determined by
   i_limits.  don't have to check for clipping here.
*/
static void
j_limits(long i, long *min_j, long *max_j)
{
  *min_j = j_min;
  *max_j = j_max;
}

/*
   returns the i, j pair corresponding to the passed position.
   returns the output position of the bin corresponding to the passed
   position.  returns 0 if the position is within bounds, non zero if
   not.
*/
static int
clip(double x_in, double y_in, CoordIndex *ci)
{
  if (x_in < x_min || x_in > x_max || y_in < y_min || y_in > y_max) return 1;
  return noclip(x_in, y_in, ci);
}

/*
   returns the i, j pair corresponding to the passed position.
   returns the output position of the bin corresponding to the passed
   position.  no clipping is done.
*/
static int
noclip(double x_in, double y_in, CoordIndex *ci)
{
  double x, y, dx, dy, xs, ys;

  /* to determine the index of the pixel in which a value falls,
     you'd like to just do
     
                x_in / x_sp

     if the pixels are aligned such that the left edges fall upon
     multiples of the pixel size, then this works.  however, that
     doesn't always occur.  instead, we move the input coordinates
     into a coordinate system where pixel edges do lie on nice numbers.
  */

  xs = x_in - c.x_off + x_shift;
  ys = y_in - c.y_off + y_shift;

  /* find index, worry about round off error */
  ci->j = floor(xs / x_sp);
  if ( (x = ci->j * x_sp) > xs)
    x = --ci->j * x_sp;
  else if ((ci->j+1) * x_sp < xs)
    x = ++ci->j * x_sp;

  x += x_center + c.x_off;

  if (fabs(dx = x - x_in) > radius)
    return 1;

  ci->i = floor(ys / y_sp);
  if ( (y = ci->i * y_sp) > ys)
    y = --ci->i * y_sp;
  else if ((ci->i+1) * y_sp < ys)
    y = ++ci->i * y_sp;

  y += y_center + c.y_off;

  /* check if outside of circle */
  if (fabs(dy = y - y_in) > radius || dy * dy + dx * dx > radius2)
    return 1;

  ci->x = x;
  ci->y = y;

  return 0;
}

/*
     returns the output coordinates of the bin corresponding to
     the passed (i,j) pair
*/
static void
bin_pos(long i, long j, CoordIndex *ci)
{
  ci->i = i;
  ci->j = j;

  ci->x = ci->j * x_sp + x_center + c.x_off;
  ci->y = ci->i * y_sp + y_center + c.y_off;
}

/*
     returns the area of the bin corresponding to the passed (i,j) pair
*/
static double
bin_area(long i, long j)
{
  return area;
}

/*
   given an i value and empiricially determined max and min j values,
   for an image, return j limits.  for rectangular grids, simply
   returning the passed limits should suffice.  for polar grids, the
   input limits have no real meaning.  when clipping is in effect,
   should return clipping limits.
*/
static void
j_image_limits(long i, long j_min_e, long j_max_e, long *min_j, long *max_j)
{
  if (c.clip)
  {
    *min_j = j_min;
    *max_j = j_max;
  }
  else
  {
    *min_j = j_min_e;
    *max_j = j_max_e;
  }
}

static void
fill_hdr(
  PixStorage pix_storage,
  long min_i,
  long max_i,
  long min_j,
  long max_j
)
{
  CoordIndex ci;

  tf_enter("rectcirc fill_header");

  bin_pos(0, 0, &ci);

  hdr.diameter = diameter;
  hdr.x_sp = x_sp;
  hdr.y_sp = y_sp;
  hdr.x0 = ci.x;
  hdr.y0 = ci.y;

  switch(pix_storage)
  {
    /* image is in core, must have been clipped */
    case CORE:
      min_i = i_min;
      max_i = i_max;
      min_j = j_min;
      max_j = j_max;
      break;

    case SPARSE:
      if (c.clip)
      {
	min_i = i_min;
	max_i = i_max;
	min_j = j_min;
	max_j = j_max;
      }
      break;

    default:
      tf_exit(ExitERR_misc, "internal error in %s at %d", __FILE__,
		      __LINE__);
  }

  hdr.i_min = min_i;
  hdr.ni    = max_i - min_i + 1;
  hdr.j_min = min_j;
  hdr.nj    = max_j - min_j + 1;

  tf_leave();
}

static void
write_header_ascii(
  PixStorage pix_storage,
  long min_i,
  long max_i,
  long min_j,
  long max_j,
  ImageDump *dump
)
{

  tf_enter("rectcirc write_header_ascii");

  fill_hdr(pix_storage, min_i, max_i, min_j, max_j);

  dump->ascii_hdr_start(0);

  dump->ascii_hdr_entry("ni",     "%d", hdr.ni);
  dump->ascii_hdr_entry("nj",	  "%d", hdr.nj);
  dump->ascii_hdr_entry("i_min",  "%d", hdr.i_min);
  dump->ascii_hdr_entry("j_min",  "%d", hdr.j_min);
  dump->ascii_hdr_entry("x0",     "%f", hdr.x0);
  dump->ascii_hdr_entry("y0",     "%f", hdr.y0);
  dump->ascii_hdr_entry("diameter",  "%f", hdr.diameter);
  dump->ascii_hdr_entry("x_sp",   "%f", hdr.x_sp);
  dump->ascii_hdr_entry("y_sp",   "%f", hdr.y_sp);

  dump->ascii_hdr_end();

  tf_leave();
}


static void
write_header_raw(
  PixStorage  pix_storage,
  long	      min_i,
  long	      max_i,
  long	      min_j,
  long	      max_j,
  ImageDump  *dump
)
{
  tf_enter("rectcirc write_header_raw");

  fill_hdr(pix_storage, min_i, max_i, min_j, max_j);

  dump->binary_hdr(rectcirc_BackEnd.type, &hdr, sizeof(hdr), hdr_map,
		   n_entries);

  tf_leave();
}


/*
     any cleanups should be done here.
     return 0 on success, non zero on failure.
*/
static int
cleanup(void)
{
  return 0;
}
