/*
 Two sets of action routines called by the redblack tree routines.
 The first set should be used when traversing the tree.  It writes
 the contents of each pixel to the output stream.

 The second set should be used whilst destroying the tree.  It frees
 the memory associated with the pixels after writing their contents to
 the output stream.

 each set has a normalized bin option as well.
*/

#include <stdlib.h>
#include <stdio.h>


#include <rbtree/rbtree.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

#include "stream_actions.h"

static void j_pos_free(void *dp);
static int j_pos_out(void *dp);
static void j_pos_free_norm(void *dp);
static int j_pos_out_norm(void *dp);

/* as both sets of routines are called by the redblack routines with a
   very limited number of arguments,  file-global variables are used to pass
   extra information to the action functions
*/

static Bpipe* bpipe;           /* where to write out, set by set_dump_stream */
static BpipePhoton* photon;                        /* set by set_dump_stream */

static long i;			/* keep track of i index for j routines */
static int normalize;		/* if true, normalize bins */

/* set dump stream for the tree dump routines */
void
set_dump_stream(Bpipe* bpipe_ptr, BpipePhoton* photon_ptr, int norm)
{
  bpipe = bpipe_ptr;
  photon = photon_ptr;
  normalize = norm;
}


/*
   this routine is called by the redblack delete routine.
   it works on an i-node. first dumping and deleting the node's
   associated j-tree.  it then deletes the i node.
*/
void
i_pos_free(void *dp)
{
  Inode *inode = (Inode *) dp;

  i = inode->i;		/* set file-global variable for i keep track */
  rbtree_delete(inode->j,
		normalize ? j_pos_free_norm : j_pos_free,
		POST_ORDER, LEFT_TO_RIGHT);
  free(inode);
}

/*
   this action routine is called by the redblack delete routine at the
   behest of i_pos_free.  it dumps out a j-node and frees it.
   it uses the tree-traversal j-node output routine to do its work.
*/
static void
j_pos_free(void *dp)
{
  Jnode *jnode = (Jnode *) dp;

  j_pos_out(dp);

  free(jnode->wt);
  free(jnode);
}


/*
   this action routine is called by the redblack delete routine at the
   behest of i_pos_free.  it dumps out a j-node and frees it.
   it uses the tree-traversal j-node output routine to do its work.
*/
static void
j_pos_free_norm(void *dp)
{
  Jnode *jnode = (Jnode *) dp;

  j_pos_out_norm(dp);

  free(jnode->wt);
  free(jnode);
}


/*
   this routine is called by the redblack traversal routine.  it works
   on an i-node. dumping the node's associated j-tree.
*/
int
i_pos_out(void *dp)
{
  Inode *inode = (Inode *) dp;

  i = inode->i;
  rbtree_traverse(inode->j,
		  normalize ? j_pos_out_norm : j_pos_out,
		  LEFT_TO_RIGHT);
  return 0;
}

/*
   this action routine is invoked by the redblack traversal routine at
   the behest of i_pos_out.  it is also called directly by j_pos_free
   to do the dirty work.  it writes the contents of the j-node to the
   output stream.
*/
static int
j_pos_out(void *dp)
{
  Jnode *jnode = (Jnode *) dp;
  CoreWeight *wt = jnode->wt;
  CoreWeight *wt_lim = jnode->wt + jnode->j_max - jnode->j_min + 1;
  long j;
  CoordIndex ci;

  for (j = jnode->j_min ; wt < wt_lim ; j++, wt++)
  {
    if (! *wt) continue;

    c.backend->bin_pos(i, j, &ci);
    *photon->x = ci.x;
    *photon->y = ci.y;
    *photon->z = 0.0;
    *photon->weight = *wt;

    if ( bpipe_write_dpkt( bpipe->bpipe, bpipe->data, bpipe->outstream ) ) {
      tf_enter("j_pos_out");

      tf_exit(ExitERR_fwrite,
		      "error writing pixel (%d, %d) to `%s'",
		      i, j, c.output);
    }
  }
  return 0;
}

/*
   this action routine is invoked by the redblack traversal routine at
   the behest of i_pos_out.  it is also called directly by j_pos_free
   to do the dirty work.  it writes the contents of the j-node to the
   output stream.
*/
static int
j_pos_out_norm(void *dp)
{
  Jnode *jnode = (Jnode *) dp;
  CoreWeight *wt = jnode->wt;
  CoreWeight *wt_lim = jnode->wt + jnode->j_max - jnode->j_min + 1;
  long j;
  CoordIndex ci;

  for (j = jnode->j_min ; wt < wt_lim ; j++, wt++)
  {
    if (! *wt) continue;

    c.backend->bin_pos(i, j, &ci);
    *photon->x = ci.x;
    *photon->y = ci.y;
    *photon->z = 0.0;
    *photon->weight = *wt / c.backend->bin_area(i, j);

    if ( bpipe_write_dpkt( bpipe->bpipe, bpipe->data, bpipe->outstream ) ) {
      tf_enter("j_pos_out");

      tf_exit(ExitERR_fwrite,
		      "error writing pixel (%d, %d) to `%s'",
		      i, j, c.output);
    }
  }
  return 0;
}

