/*
   wedge back end routines.  global information is available via
   the program common structure, c
*/

#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include <mathconst/mathconst.h>
#include <tparam/tparam.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"
#include "wedge.h"

/*************************************************************************/

/* prototypes for statically defined functions in this file */

static void gparam_rl( paramfile pf );
static int setup( void );
static int i_limits( long *min_i, long *max_i, unsigned long *nbins );
static void j_limits( long i, long *min_j, long *max_j );
static int clip( double x_in, double y_in, CoordIndex *ci );
static int noclip( double x_in, double y_in, CoordIndex *ci );
static void bin_pos( long i, long j, CoordIndex *ci );
static double bin_area( long i, long j );
static void j_image_limits( long i, long j_min_e, long j_max_e, long *min_j,
			    long *max_j );
static void write_header_ascii( PixStorage pix_storage, long min_i, long max_i,
				long min_j, long max_j, ImageDump *dump );
static void write_header_raw( PixStorage pix_storage, long min_i, long max_i,
			      long min_j, long max_j, ImageDump *dump );
static int cleanup( void );
static int find_list_rdb_wedge( double radius_sqr, double theta,
				CoordIndex *ci );
static int find_linear( double x, double y, CoordIndex *ci );
static int find_exp( double x, double y, CoordIndex *ci );



/*************************************************************************/

BackEnd wedge_BackEnd =
{
  "wedge",			/* back end's name */
  BackEnd_Wedge,		/* back end's id */
  gparam_rl,			/* parameter read read routine */
  setup,			/* setup routine */
  i_limits,			/* return primary index limits for radial R */
  j_limits,			/* return secondary index limits for theta */
  noclip,			/* quantize routine (set by setup) */
  bin_pos,			/* return real coordinates from indices */
  bin_area,			/* return area of bin from indices */
  j_image_limits,		/* empirical secondary index limits */
  1,				/* front end should determine actual
  				   min and max secondary coordinate
  				   indices */
  write_header_raw,		/* write raw image header */
  write_header_ascii,		/* write ascii image header */
  cleanup			/* cleanup routine */
};

/*************************************************************************/

/* the output header */
static WedgeImage hdr;
#define MapEntry(entry) {offsetof(WedgeImage, entry), sizeof(hdr.entry)}

static HdrMapEntry hdr_map[] = {
  MapEntry(type),
  MapEntry(Nwedge),
  MapEntry(constant),
  MapEntry(base),
  MapEntry(Rmax),
  MapEntry(Tmin),
};
#define n_entries (sizeof(hdr_map) / sizeof(HdrMapEntry))

/* external parameters, filled in by gparam */
/* Global variables, you just got to love `em */
static Annuli*       annuli_list_rdb=NULL;
static AnnuliExp*    annuli_exp=NULL;
static AnnuliLinear* annuli_linear=NULL;
/*
don't ask me why the following does not work:
#include <limits.h>
static double rclip=DBL_MAX;
*/
static double        rclip=1.0e100;
/*************************************************************************/

static void gparam_rl( paramfile pf ) {

  char rtype[ NAME_LENGTH + 1 ];
  char rspec[ NAME_LENGTH + 1 ];

  tf_enter("gparam wedge");

  tpgetstr( pf, "rtype", rtype, NAME_LENGTH, 1 );
  tpgetstr( pf, "rspec", rspec, NAME_LENGTH, 1 );
  rclip = tpgetd( pf, "n_x" );

  /*
    If rtype is equal to file or discrete then
    one must allocate memory to store the annuli.
   */
  if ( 0 == strcmp( rtype, "discrete" ) ) {
    /* The finite-discrete set of wedges */
    annuli_list_rdb = read_wedge_list( rspec );
    c.clip = 1;
  } else if ( 0 == strcmp( rtype, "file" ) ) {
    /* The rdb file contains the descriptions for the annuli */
    annuli_list_rdb = read_wedge_rdb_file( rspec );
    c.clip = 1;
  } else if ( 0 == strcmp( rtype, "exp" ) )
    /* an `infinite' exponential wedge */
    annuli_exp = read_wedge_exp( rspec );
  else if ( 0 == strcmp( rtype, "linear" ) )
    /* an `infinite' linear wedge */
    annuli_linear = read_wedge_linear( rspec );

  tf_leave();
}

/*
     any initializations should be done here.
     return 0 on success, non zero on failure.
     note that the radial R is indexed by i 
     and theta is indexed by j.
*/
static int setup( void ) {

  if ( (NULL != annuli_exp) || (NULL != annuli_linear && c.clip) ) 
  {


  }

  if ( c.clip ) {
    wedge_BackEnd.quantize = clip;
  } else {
    wedge_BackEnd.quantize = noclip;
  }

  return 0;

}


/*
   returns the minimum and maximum bin index in the i dimension, as well
   as the total number of bins. returns 0 if all information is
   available, 1 if any is missing
*/
static int i_limits( long *min_i, long *max_i, unsigned long *nbins ) {

  /* only know limits if clipping is in effect */
  if ( c.clip ) {

    *min_i = 0;
    *max_i = annuli_list_rdb->num_annulus - 1;

    {
      int i;
      for ( i = 0, *nbins = 0; i < annuli_list_rdb->num_annulus; i++ )
	*nbins += annuli_list_rdb->annulus[ i ].num_wedges;
    }

  } else
    return 1;

  return 0;
}


/*
   for a given i, return the minimum and maximum j values.  guaranteed
   to be called only with legal i values as determined by
   i_limits.  don't have to check for clipping here.
*/
static void j_limits( long i, long *min_j, long *max_j ) {
  *min_j = 0;

  if ( annuli_list_rdb ) {
    *max_j = annuli_list_rdb->annulus[ i ].num_wedges - 1;
  } else if ( annuli_exp ) {
    *max_j = annuli_exp->num_wedges - 1;
  } else
    *max_j = annuli_linear->num_wedges - 1;

}

/*
   returns the i, j pair corresponding to the passed position.
   returns the output position of the bin corresponding to the passed
   position.  returns 0 if the position is within bounds, non zero if
   not.
*/
static int clip( double x_in, double y_in, CoordIndex *ci ) {

  if ( x_in * x_in + y_in * y_in > rclip )
    return NotWithinWedge;

  return noclip( x_in, y_in, ci );

}

/*
  PROCEDURE
  The system's atan2 returns a value from -Pi to Pi,
  so it is more convenient to work in degrees.

  RETURN
  The angle in degrees.
 */
static double myatan2( double x, double y ) {

  double result = 180.0 * atan2( y, x ) / M_PI;

  if ( result < 0.0 )
    result += 360.0;

  return result;

}

/*
   returns the i, j pair corresponding to the passed position.
   returns the output position of the bin corresponding to the
   passed position.  no clipping is done.
*/
static int noclip( double x_in, double y_in, CoordIndex *ci ) {

  if ( annuli_list_rdb )
    return find_list_rdb_wedge( x_in, y_in, ci );
      
  else if ( annuli_exp )
    return find_exp( x_in, y_in, ci );

  else
    return find_linear( x_in, y_in, ci );

}

/*
     returns the output coordinates of the bin
     corresponding to the passed (i,j) pair
*/
static void bin_pos( long i, long j, CoordIndex *ci ) {

  double rmax=0.0, rmin=0.0, theta_max, theta_min;
  double r, theta, theta_radians;

  ci->i = i;
  ci->j = j;

  if ( annuli_list_rdb ) {
    rmax = annuli_list_rdb->annulus[ i ].r_max;
    rmin = annuli_list_rdb->annulus[ i ].r_min;
    theta_max = annuli_list_rdb->annulus[ i ].wedge[ j ].theta_max;
    theta_min = annuli_list_rdb->annulus[ i ].wedge[ j ].theta_min;
  } else if ( annuli_exp ) {
    rmax = annuli_exp->constant * pow( annuli_exp->base, (double) i );
    if ( 0 != i )
      rmin = annuli_exp->constant * pow( annuli_exp->base, (double) i - 1 );
    else
      rmin = 0.0;
    theta_max = annuli_exp->wedge[ j ].theta_max;
    theta_min = annuli_exp->wedge[ j ].theta_min;
  } else {
    rmax = i * annuli_linear->deltaR;
    rmin = rmax - annuli_linear->deltaR;
    theta_max = annuli_linear->wedge[ j ].theta_max;
    theta_min = annuli_linear->wedge[ j ].theta_min;
  }

  r = ( rmax + rmin ) / 2.0;
  theta = ( theta_max + theta_min ) / 2.0;
  theta_radians = M_PI * theta / 180.0;
  ci->x = r * cos( theta_radians ) + c.x_off;
  ci->y = r * sin( theta_radians ) + c.y_off;

}

/*
     returns the area of the bin corresponding to the passed (i,j) pair
*/
static double bin_area( long i, long j ) {

  double rmax_sqr=0.0, rmin_sqr=0.0;
  int num_wedges=0;

  if ( annuli_list_rdb ) {
    rmax_sqr = annuli_list_rdb->annulus[ i ].r_max_sqr;
    rmin_sqr = annuli_list_rdb->annulus[ i ].r_min_sqr;
    num_wedges = annuli_list_rdb->annulus[ i ].num_wedges;
  } else if ( annuli_exp ) {
    double rmin=0.0;
    double rmax = annuli_exp->constant * pow( annuli_exp->base, (double) i );
    if ( 0 != i )
      rmin = annuli_exp->constant * pow( annuli_exp->base, (double) i - 1 );
    rmax_sqr = rmax * rmax;
    rmin_sqr = rmin * rmin;
  } else {
    double rmax = i * annuli_linear->deltaR;
    double rmin = rmax - annuli_linear->deltaR;
    rmax_sqr = rmax * rmax;
    rmin_sqr = rmin * rmin;
  }

  return M_PI * ( rmax_sqr - rmin_sqr ) * ( 360.0 / num_wedges );

}

/*
   given an i value and empiricially determined max and min j values,
   for an image, return j limits.  for rectangular grids, simply
   returning the passed limits should suffice.  for polar grids, the
   input limits have no real meaning.  when clipping is in effect,
   should return clipping limits.
*/
static void j_image_limits( long i, long j_min_e, long j_max_e, long *min_j,
			    long *max_j ) {

  /*
  ?????????????????????????????????????????????????????????????????????????????
  if ( c.clip ) {
    *min_j = j_min;
    *max_j = j_max;
  } else {
    *min_j = j_min_e;
    *max_j = j_max_e;
  }
  ?????????????????????????????????????????????????????????????????????????????
  */
}

static void fill_hdr( PixStorage pix_storage, long min_i, long max_i,
		      long min_j, long max_j ) {

  CoordIndex ci;

  tf_enter( "wedge fill_hdr");

#ifdef FOOBAR

  bin_pos(0, 0, &ci);

  hdr.x_sz = x_sz;
  hdr.y_sz = y_sz;
  hdr.x0 = ci.x;
  hdr.y0 = ci.y;

  switch(pix_storage)
  {
    /* image is in core, must have been clipped */
    case CORE:
      min_i = i_min;
      max_i = i_max;
      min_j = j_min;
      max_j = j_max;
      break;

    case SPARSE:
      if (c.clip)
      {
	min_i = i_min;
	max_i = i_max;
	min_j = j_min;
	max_j = j_max;
      }
      break;

    default:
      tf_exit(ExitERR_misc, "internal error in %s at %d", __FILE__,
		      __LINE__);
  }

  hdr.i_min = min_i;
  hdr.ni    = max_i - min_i + 1;
  hdr.j_min = min_j;
  hdr.nj    = max_j - min_j + 1;

#endif

  tf_leave();
}

static void
write_header_ascii(
  PixStorage pix_storage,
  long min_i,
  long max_i,
  long min_j,
  long max_j,
  ImageDump *dump
)
{

  tf_enter("rectrect write_header_ascii");

  fill_hdr(pix_storage, min_i, max_i, min_j, max_j);

  dump->ascii_hdr_start(0);

  dump->ascii_hdr_entry( "type",     "%d", hdr.type );
  dump->ascii_hdr_entry( "Nwedge",   "%d", hdr.Nwedge );
  dump->ascii_hdr_entry( "constant", "%f", hdr.constant );
  dump->ascii_hdr_entry( "base",     "%f", hdr.base );
  dump->ascii_hdr_entry( "Rmax",     "%f", hdr.Rmax );
  dump->ascii_hdr_entry( "Tmin",     "%f", hdr.Tmin );

  dump->ascii_hdr_end();

  tf_leave();
}


static void
write_header_raw(
  PixStorage  pix_storage,
  long	      min_i,
  long	      max_i,
  long	      min_j,
  long	      max_j,
  ImageDump  *dump
)
{
  tf_enter("rectrect write_header_raw");

  fill_hdr(pix_storage, min_i, max_i, min_j, max_j);

  dump->binary_hdr(wedge_BackEnd.type, &hdr, sizeof(hdr), hdr_map,
		   n_entries);

  tf_leave();
}

static void
free_Annuli( Annuli* ptr ) {

  if ( ptr ) {
    int i;
    for ( i = 0; i < ptr->num_annulus; i++ )
      free( ptr->annulus[ i ].wedge );
    ptr->num_annulus=0;
    free( ptr->annulus );
    free( ptr );
  }

}

static void
free_AnnuliExp( AnnuliExp* ptr ) {

  if ( ptr ) {
    ptr->num_wedges=0;
    free( ptr->wedge );
    free( ptr );
  }

}

static void
free_AnnuliLinear( AnnuliLinear* ptr ) {

  if ( ptr ) {
    ptr->num_wedges=0;
    free( ptr->wedge );
    free( ptr );
  }

}

/*
     any cleanups should be done here.
     return 0 on success, non zero on failure.
*/
static int
cleanup( void ) {

  free_Annuli( annuli_list_rdb );
  free_AnnuliExp( annuli_exp );
  free_AnnuliLinear( annuli_linear );

  return 0;
}

/*
  PROCEDURE
    Calculates the min-max angle of a wedge.

  RETURN
 */
void calculate_wedges( Wedge** wedge_ptr, char* tmin_degree, int num_wedges ) {

  int j;

  double delta_theta=0.0;
  double tmin=0.0;

  delta_theta = 360.0 / num_wedges;

  tf_enter( "calculate_wedges" );

  if ( EXIT_FAILURE == str2d( tmin_degree, &tmin ) )
    tf_exit( ExitERR_illegalArg, "Illegal entry" );

  *wedge_ptr =  (Wedge*) malloc( num_wedges * sizeof( Wedge ) );
    if ( NULL == *wedge_ptr )
      tf_exit(  ExitERR_alloc, "Unable to allocate memory for wedge_ptr" );

  for ( j = 0; j < num_wedges; j++ ) {

    double theta_min = tmin + j * delta_theta;
    double theta_max = theta_min + delta_theta;

    (*wedge_ptr)[ j ].theta_min = theta_min;
    (*wedge_ptr)[ j ].theta_max = theta_max;

    /*
      fprintf( stderr, "calculate_wedges( ) : theta_min = %g\t", theta_min );
      fprintf( stderr, "theta_max = %g\n", theta_max );
    */

  }

  tf_leave( );

}

/*
  PROCEDURE
  convert a string, txt, to a double, result.

  RETURN
  EXIT_SUCCESS or EXIT_FAILURE.  If EXIT_SUCCESS is returned
  then argument result holds the value of the string txt.
 */
int str2d( char* txt, double* result ) {

  char* ptr = NULL;

  tf_enter( "str2d" );

  *result = 0;

  /* Field is empty */
  if ( NULL == txt || '\0' == *txt ) {
    tf_message( "Empty field" );
    tf_leave( );
    return EXIT_FAILURE;
  }

  *result = strtod( txt, &ptr );

  if ( ERANGE == errno ) {
    *result = 0;
    tf_message( "Variable overflow `%s'", txt );
    tf_leave( );
    return EXIT_FAILURE;
  }

  if ( '\0' != *ptr ) {
    *result = 0;
    tf_message( "Not a number, problem at `%s'", ptr );
    tf_leave( );
    return EXIT_FAILURE;
  }

  /* contains a legit number */
  return EXIT_SUCCESS;

}

/*
  PROCEDURE
  convert a string, txt, to an int, result.

  RETURN
  EXIT_SUCCESS or EXIT_FAILURE.  If EXIT_SUCCESS is returned
  then argument result holds the value of the string txt.
 */
int str2i( char* txt, int* result ) {

  double tmp;

  tf_enter( "str2i" );

  if ( EXIT_FAILURE == str2d( txt, &tmp ) ) {
    tf_leave( );
    return EXIT_FAILURE;
  }

  *result = atoi( txt );

  tf_leave( );

  return EXIT_SUCCESS;

}

static void print_Wedge( FILE* fp, Wedge* wedge ) {

  fprintf( fp, "  Wedge( ) : theta_min = %g, theta_max = %g",
	   wedge->theta_min, wedge->theta_max );
  
}

static void print_Annulus( FILE* fp, Annulus* ptr ) {

  int i;
  fprintf( fp, 
	   " Annulus( ) : r_min_sqr = %g\tr_max_sqr = %g\tnum_wedges = %d\n",
	   ptr->r_min_sqr, ptr->r_max_sqr, ptr->num_wedges );
  for ( i = 0; i < ptr->num_wedges; i++ ) {
    print_Wedge( fp, &(ptr->wedge[ i ]) );
    fprintf( fp, "\n" );
  }

}

void print_Annuli( FILE* fp, Annuli* ptr ) {

  int i;
  fprintf( fp, "print_Annuli( ) : ptr->num_annulus = %d\n", ptr->num_annulus );
  for ( i = 0; i < ptr->num_annulus; i++ ) {
    print_Annulus( fp, &(ptr->annulus[ i ]) );
    fprintf( fp, "\n" );
  }

}

void print_AnnuliExp( FILE* fp, AnnuliExp* ptr ) {

  int i;

  fprintf( fp, "print_AnnuliExp( ) : base=%g, constant=%g, log10base=%g,"
	   "num_wedges=%d\n",
	   ptr->base, ptr->constant, ptr->lnbase, ptr->num_wedges );
  
  for ( i = 0; i < ptr->num_wedges; i++ ) {
    print_Wedge( fp, &(ptr->wedge[ i ]) );
    fprintf( fp, "\n" );
  }

}

void print_AnnuliLinear( FILE* fp, AnnuliLinear* ptr ) {

  int i;

  fprintf( fp, "print_AnnuliLinear( ) : deltaR=%g, num_wedges=%d\n",
	   ptr->deltaR, ptr->num_wedges );

  for ( i = 0; i < ptr->num_wedges; i++ ) {
    print_Wedge( fp, &(ptr->wedge[ i ]) );
    fprintf( fp, "\n" );
  }

}

static void free_Annulus( Annulus* ptr ) {

  if ( ptr )
    free( ptr );

}


int cmp_annulus( const void* k, const void* datum ) {

  double* key = (double*) k;
  Annulus* annulus = (Annulus*) datum;

  if ( *key < annulus->r_min_sqr )
    return -1;
  else if ( annulus->r_max_sqr < *key )
    return 1;
  else
    return 0;

}

int cmp_Wedge_theta( const void* k, const void* datum ) {

  double* key = (double*) k;
  Wedge* wedge = (Wedge*) datum;

  if ( *key < wedge->theta_min )
    return -1;
  else if ( wedge->theta_max < *key )
    return 1;
  else
    return 0;
  
}

static int find_list_rdb_wedge( double x, double y, CoordIndex *ci ) {

  double radius_sqr = x * x + y * y;

  Annulus* annulus = 
    bsearch( (const void*) &radius_sqr, (const void*) annuli_list_rdb->annulus,
	     annuli_list_rdb->num_annulus, sizeof( Annulus ), cmp_annulus );

  /* The photon did not fall in any of the radial annuli, 
     tell the user by returning a non-zero value */
  if ( NULL == annulus )
    return NotWithinWedge;

  {

    double theta = myatan2( x, y );
    Wedge* wedge = 
      bsearch( (const void*) &theta, (const void*) annulus->wedge,
	       annulus->num_wedges, sizeof( Wedge ), cmp_Wedge_theta );

    if ( NULL == wedge )
      return NotWithinWedge;

    bin_pos( annulus - annuli_list_rdb->annulus, wedge - annulus->wedge, ci );

  }

  return WithinWedge;

}

static int find_exp( double x, double y, CoordIndex *ci ) {

  int radial_i;

  /*
               n - 1                            n
    Rmin = a b         <= radius  <  Rmax = a b

  */

  double radius = sqrt( x * x + y * y );
  if ( radius < annuli_exp->constant )
    radial_i = 0;
  else {
    radial_i = 
      ceil( log( radius / annuli_exp->constant ) / annuli_exp->lnbase );
  }

  {
    double theta = myatan2( x, y );

    Wedge* wedge = NULL;

    wedge = 
      bsearch( (const void*) &theta, (const void*) annuli_exp->wedge,
	       annuli_exp->num_wedges, sizeof( Wedge ), cmp_Wedge_theta );

    if ( NULL == wedge )
      return NotWithinWedge;

    /*
    fprintf( stderr, "n=%d\t%g\t%g\t%g\n", radial_i,
	     annuli_exp->constant * pow( annuli_exp->base, radial_i - 1 ),
	     radius,
	     annuli_exp->constant * pow( annuli_exp->base, radial_i ) );
	     */

    bin_pos( radial_i, wedge - annuli_exp->wedge, ci );

  }

  return WithinWedge;

}

static int find_linear( double x, double y, CoordIndex *ci ) {

  int radial_i;

  /*
    Rmin = n * deltaR <= radius < Rmax = ( n + 1 ) * deltaR;
  */
  double radius = sqrt( x * x + y * y );
  double tmp = radius / annuli_linear->deltaR;
  radial_i = tmp;

  {
    double theta = myatan2( x, y );
    Wedge* wedge = 
      bsearch( (const void*) &theta, (const void*) annuli_linear->wedge,
	       annuli_linear->num_wedges, sizeof( Wedge ), cmp_Wedge_theta );

    if ( NULL == wedge )
      return NotWithinWedge;

    ci->j = wedge - annuli_linear->wedge;

    bin_pos( radial_i, wedge - annuli_linear->wedge, ci );
  }

  return WithinWedge;

}
