#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <tracefct/exiterrvals.h>
#include <tparam/tparam.h>
#include <suplib/str.h>
#include <tracefct/tracefct.h>

#include "rwbpipe.h"
#include "wedge.h"

/*
  PROCEDURE
    The function creates an Annuli structure to hold the list of annuli.
    The input string, str, is expected to hold a string of the form:
         Rmax,Nwedges,Tmin Rmax,Nwedges,Tmin ... Rmax,Nwedges,Tmin
    The delimiters for the lists are slash or colon (/ or :).
    The delimters within the lists are comma or space (, or ' ')

  RETURN
    The structure Annuli to store the wedges.
 */
Annuli* read_wedge_list( char str[] ) {

  char** wedge_tokens=NULL;
  int num_annuli=0, i;
  Annuli* annuli=NULL;

  tf_enter( "read_wedge_list" );

  annuli = (Annuli*) malloc( sizeof( Annuli ) );
  if ( NULL == annuli )
    tf_exit( ExitERR_alloc, "Unable to allocate memory for annuli" );

  /*
    Tokenize the list of annuli:

       R    , N       , Theta    /  .....  :R      N         Theta    
        max    wedges        min             max    wedges        min

    The delimiter can be either slash, space or colon (`/', ` ', or `:'
    respectively)
  */
  annuli->num_annulus = str_tokenize( str, "/: ", &wedge_tokens, 1, 1 );
  annuli->annulus =
    (Annulus*) malloc( annuli->num_annulus * sizeof( Annulus ) );
  if ( NULL == annuli->annulus )
    tf_exit( ExitERR_alloc, "Unable to allocate memory for annulus" );

  for ( i = 0; i < annuli->num_annulus; i++ ) {

    char** annulus_tokens=NULL;

    /*
      Tokenize the individual annulus:

       R    , N       , Theta
        max    wedges        min

      The delimiter is a comma (,).
    */
    int three = str_tokenize( wedge_tokens[ i ], ",", &annulus_tokens, 1, 1 );

    if ( 3 != three )
      tf_exit( ExitERR_illegalArg, "Illegal entry `%s'", wedge_tokens[ i ] );

    if ( 0 == i )
      annuli->annulus[ i ].r_min = 0;
    else
      annuli->annulus[ i ].r_min = annuli->annulus[ i - 1 ].r_max;

    if ( EXIT_FAILURE == 
	 str2d( annulus_tokens[ 0 ], &(annuli->annulus[ i ].r_max) ) )
      tf_exit( ExitERR_illegalArg, "Illegal entry" );

    /* make sure that rmin < rmax for each annulus */
    if ( annuli->annulus[ i ].r_min > annuli->annulus[ i ].r_max )
      tf_exit( ExitERR_illegalArg,
	       "Rmin  (r_min[%d]=%g) must be > Rmax (r_max[%d]=%g)",
	       i, annuli->annulus[ i ].r_min, i,
	       annuli->annulus[ i ].r_max );
    if ( EXIT_FAILURE == 
	 str2i( annulus_tokens[ 1 ], &(annuli->annulus[ i ].num_wedges) ) )
      tf_exit( ExitERR_illegalArg, "Illegal entry" );

    calculate_wedges( &(annuli->annulus[ i ].wedge), annulus_tokens[ 2 ],
		      annuli->annulus[ i ].num_wedges );

    str_tokenize_free( annulus_tokens, 1 );
     
  }

  str_tokenize_free( wedge_tokens, 1 );

  /*
    let's square the radius, this will save some computations
    when searching for the wedge where the photon lands in.
  */
  {
    int j;
    for ( j = 0; j < annuli->num_annulus; j++ ) {
      annuli->annulus[ j ].r_min_sqr = 
	annuli->annulus[ j ].r_min * annuli->annulus[ j ].r_min;
      annuli->annulus[ j ].r_max_sqr =
	annuli->annulus[ j ].r_max * annuli->annulus[ j ].r_max;
    }
  }

  /* To see the annuli-wedges etc.. uncomment this:
     print_Annuli( stderr, annuli ); */

  tf_leave( );

  return annuli;

}

