#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <tracefct/exiterrvals.h>
#include <mst_rdb/mst_rdb.h>
#include <suplib/str.h>
#include <tracefct/tracefct.h>

#include "wedge.h"

/* A local struct to be used to call the mst_rdb library */
typedef struct {
  double Rmin, Rmax, Tmin, Tmax;
} WedgeRdb;

static int cmp_Rmin( const void* a, const void* b ) {

  WedgeRdb* aa = (WedgeRdb*) a;
  WedgeRdb* bb = (WedgeRdb*) b;

  return aa->Rmin - bb->Rmin;

}

static int cmp_Tmin( const void* a, const void* b ) {

  Wedge* aa = (Wedge*) a;
  Wedge* bb = (Wedge*) b;

  return aa->theta_min - bb->theta_min;

}

/*
  PROCEDURE
  Determine the number of wedges within the ring

  RETURN
  The number of wedges which have the same wedge_rdb[ start ].Rmin
 */
static int determine_num_wedges( WedgeRdb* wedge_rdb, long start, long end ) {

  int i;
  int num_wedges=0;
  double rmin = wedge_rdb[ start ].Rmin;

  for ( i = start; i < end; i++ ) {
    if ( rmin != wedge_rdb[ i ].Rmin )
      return num_wedges;
    num_wedges++;
  }

  return num_wedges;

}

/*
  PROCEDURE
  Determine the number of rings which have different Rmin values.

  RETURN
  The number of distinct rings.
 */
static int determine_num_annulus( WedgeRdb* wedge_rdb, long num_wedges ) {

  int i;
  int num_annulus;
  double rmin = wedge_rdb[ 0 ].Rmin;

  for ( i = 0, num_annulus = 1; i < num_wedges; i++ ) {

    if ( rmin != wedge_rdb[ i ].Rmin ) {
      rmin = wedge_rdb[ i ].Rmin;
      num_annulus++;
    }

  }

  return num_annulus;

}

/*
  PROCEDURE

  RETURN
 */
static Annuli* make_annuli( WedgeRdb* wedge_rdb, long num_wedges ) {

  int i;
  Annuli* annuli = NULL;

  tf_enter( "make_annuli" );

  /* Make sure wedge_rdb is sorted by Rmin */
  qsort( wedge_rdb, num_wedges, sizeof( WedgeRdb ), cmp_Rmin );

  annuli = (Annuli*) malloc( sizeof( Annuli ) );
  if ( NULL == annuli )
    tf_exit( ExitERR_alloc, "Unable to allocate memory for annuli" );

  /* Determine the number of annulus rings */
  annuli->num_annulus = determine_num_annulus( wedge_rdb, num_wedges );

  annuli->annulus =
    (Annulus*) malloc( annuli->num_annulus * sizeof( Annulus ) );
  if ( NULL == annuli->annulus )
    tf_exit( ExitERR_alloc, "Unable to allocate memory for annulus" );

  {

    int start = 0;

    for ( i = 0; i < annuli->num_annulus; i++ ) {

      int j;
      Wedge* wedge_ptr = NULL;

      annuli->annulus[ i ].num_wedges =
	determine_num_wedges( wedge_rdb, start, num_wedges );

      annuli->annulus[ i ].wedge = 
	(Wedge*) malloc( annuli->annulus[ i ].num_wedges * sizeof( Wedge ) );
      if ( NULL == annuli->annulus[ i ].wedge )
	tf_exit(  ExitERR_alloc, "Unable to allocate memory for wedge" );

      annuli->annulus[ i ].r_min = wedge_rdb[ start ].Rmin;
      annuli->annulus[ i ].r_min_sqr = 
	wedge_rdb[ start ].Rmin * wedge_rdb[ start ].Rmin;

      annuli->annulus[ i ].r_max = wedge_rdb[ start ].Rmax;
      annuli->annulus[ i ].r_max_sqr =
	wedge_rdb[ start ].Rmax * wedge_rdb[ start ].Rmax;

      wedge_ptr = annuli->annulus[ i ].wedge;

      for ( j = 0; j < annuli->annulus[ i ].num_wedges; j++ ) {

	wedge_ptr[ j ].theta_min = wedge_rdb[ start + j ].Tmin;
	wedge_ptr[ j ].theta_max = wedge_rdb[ start + j ].Tmax;

      }

      /* Make sure the wedges within the ring are sorted in theta */
      qsort( wedge_ptr, annuli->annulus[ i ].num_wedges, sizeof( Wedge ),
	     cmp_Tmin );

      start += annuli->annulus[ i ].num_wedges;

    }

  }

  /* To see the annuli-wedges etc.. uncomment this:
     print_Annuli( stderr, annuli ); */

  tf_leave( );

  return annuli;

}

/*
  PROCEDURE
    Read the rdb file, filename, to read  the 
 */
Annuli* read_wedge_rdb_file( char filename[] ) {
  
  Annuli* annuli=NULL;
  FILE* infile=NULL;
  long num_wedges=0;
  WedgeRdb* wedge_rdb=NULL;
  rdbHeader* head=NULL;
  DataColumnMap_st* map=NULL;

  tf_enter( "read_wedge_rdb_file" );

  infile = fopen( (char*) filename, "r" );
  if ( NULL == infile )
    tf_exit( ExitERR_illegalArg, "Unable to open file `%s'", filename );

  head = rdb_rd_hdr( infile );
  if ( NULL == head ) {
    fclose( infile );
    tf_exit( ExitERR_illegalArg, "NULL == head" );
  }

  {
    RDBFieldStInfo fields[]={
      RDBentry( Rmin,      RDB_Num,   WedgeRdb ),
      RDBentry( Rmax,      RDB_Num,   WedgeRdb ),
      RDBentry( Tmin,      RDB_Num,   WedgeRdb ),
      RDBentry( Tmax,      RDB_Num,   WedgeRdb ),
    };

    map = rdb_map_cols_stst( head, sizeof(fields) / sizeof(RDBFieldStInfo),
			     fields );
    if ( NULL == map ) {
      rdb_free_hdr( head );
      fclose( infile );
      tf_exit( ExitERR_illegalArg, "NULL == map" );
    }
  }

  num_wedges = rdb_count( infile, head );
  if ( num_wedges < 0 ) {
    rdb_free_hdr( head );
    rdb_free_map( map );
    fclose( infile );
    tf_exit( ExitERR_illegalArg, "num_wedges = %d", num_wedges );
  }

  wedge_rdb = (WedgeRdb*) malloc( sizeof( WedgeRdb ) * num_wedges );
  if ( NULL == wedge_rdb ) {
    rdb_free_hdr( head );
    rdb_free_map( map );
    fclose( infile );
    tf_exit( ExitERR_alloc, "Unable to allocate memory for wedge_rdb" );
  }

  {
    int i = 0;
    while( rdb_col_read_st( infile, head, map, &wedge_rdb[ i ] ) ) {
      /*
      fprintf( stderr, "%g\t%g\t%g\t%g\n",
	       wedge_rdb[ i ].Rmin, wedge_rdb[ i ].Rmax,
	       wedge_rdb[ i ].Tmin, wedge_rdb[ i ].Tmax );
      */
      i++;
    }
  }

  annuli = make_annuli( wedge_rdb, num_wedges );
  free( wedge_rdb );

  rdb_free_hdr( head );
  rdb_free_map( map );
  fclose( infile );

  tf_leave( );

  return annuli;

}
