// --8<--8<--8<--8<--
//
// Copyright (C) 2013 Smithsonian Astrophysical Observatory
//
// This file is part of wedge
//
// wedge is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// -->8-->8-->8-->8--

#include <cmath>

#include <mathconst/mathconst.h>

#include <tracefctxx/TraceFct.h>

#include <bpipe/bpipe.h>
#include <bpipe/datatypes.h>
#include <bpipe/bpipe_axaf.h>

#include "process.h"
#include "param.h"

/* Routine to get the data pointer for a named data packet field */

static char *
dpktf_data( BPipe *bpipe, char *name, void *data ) {

    TraceFct tf( "dpktf_data" );
    DpktField *field = bpipe_dpktf( bpipe, name );

    if ( field == NULL )
	tf.die( "%s isn't in input data stream", name);

    return bpipe_dpktf_data( field, data );

}

void
process(Pars& pars) {

    TraceFct tf( "process" );
    BPipe *bpipe;
    BPipeOutput *bpo;
    void  *data;
    size_t nphot;
    DVector3 *position;
    DVector3 *direction;

    int sign_min, sign_max;
    double xmin, xmax, ymin, ymax, xtest, ytest, mid_theta;


    /* Setup bpipe input and output streams */

    if ( ( bpipe = bpipe_new() ) == NULL)
	tf.die( "couldn't open bpipe: %s",
		bpipe_strerror(bpipe_errno));

    if ( ( bpipe_input(bpipe, pars.input.c_str()) ) != NULL)
	tf.die( "couldn't open input file `%s': %s", pars.input.c_str(),
		bpipe_strerror(bpipe_errno));

    if ( NULL == ( bpo = bpipe_output( bpipe, pars.output.c_str() ) ) )
	tf.die( "couldn't open input file `%s': %s", pars.output.c_str(),
		bpipe_strerror(bpipe_errno));

    if (NULL == (data = bpipe_map_alloc(bpipe,1, NULL)))
	tf.die( "unable to allocate memory for data packet: %s",
		 bpipe_strerror(bpipe_errno));

    position   = (DVector3 *) dpktf_data( bpipe, BP_AXAF_phot_pos, data);
    direction  = (DVector3 *) dpktf_data( bpipe, BP_AXAF_phot_dir, data);

    if ( bpipe_write_hdr(bpipe) )
	tf.die( "bpipe_write_hdr: %s",bpipe_strerror(bpipe_errno));

    /* this routine works by creating two lines which bound the wedge's
       edges.  the lines split the plane into four sections.  points in
       each section are uniquely tagged by a pair of numbers which
       denote on which side of the two lines that the points lie.  if the
       wedge is less than 180 degrees in width, it occupies one of the
       four sections.  if it is greater than 180 degrees, it occupies
       three of the four sections.  the algorithm determines if a point
       is in the wedge by constructing a test point which lies in the
       inclusive section (if the wedge is less than 180 degrees) or the
       exclusive section(if the wedge is greater than 180 degrees) and
       comparing the unique id pair of the test point with that of the
       ray.
    */

    /* generate comparison points for the angles */

    /* remove 360 degree degeneracy and turn into radians */

    pars.theta_min = fmod(pars.theta_min, 360.0) * M_PI / 180.0;
    pars.theta_max = fmod(pars.theta_max, 360.0) * M_PI / 180.0;

    /* if the wedge straddles the discontinuity, make it continuous */
    if (pars.theta_min > pars.theta_max) pars.theta_max += M_2PI;

    /* if the wedge is greater than 180 degrees in width, make the
       exclusive section the wedge and flip the notion of rejection and
       acceptance
    */

    if (pars.theta_max - pars.theta_min > M_PI)
    {
	double temp = pars.theta_max;
	pars.theta_max = pars.theta_min;
	pars.theta_min = temp;
	pars.reject = !pars.reject;
	/* ensure that theta_min < theta_max */
	pars.theta_max += M_2PI;
    }

    /* determine the direction cosines of the two lines */

    xmin = cos(pars.theta_min);
    ymin = sin(pars.theta_min);

    xmax = cos(pars.theta_max);
    ymax = sin(pars.theta_max);

    /* construct test point exactly half way between edges of wedge */

    mid_theta = (pars.theta_min + pars.theta_max) / 2.0;
    xtest = cos(mid_theta);
    ytest = sin(mid_theta);

    /* determine which side of the lines the test point is */

    sign_min = (xtest * ymin - ytest * xmin) > 0.0;
    sign_max = (xtest * ymax - ytest * xmax) > 0.0;

    /* read-in bpipe photon data packets */

    bpipe_errno = BPNOERROR;
    nphot = 0;

    while( bpipe_read_dpkts( bpipe, data, 1 ) == 1) {

	double t;
	double x, y;

	nphot++;

	t = (pars.z - position->z) / direction->z;
	x = position->x + t * direction->x - pars.x;
	y = position->y + t * direction->y - pars.y;

	bool in_wedge =
	    ((x * ymin) - (y * xmin) > 0.0) == sign_min &&
	    ((x * ymax) - (y * xmax) > 0.0) == sign_max;

	if (    ( in_wedge && pars.reject )
	     || ( !in_wedge && ! pars.reject )
	   )
	    continue;

	if ( bpipe_write_dpkt( bpipe, data, bpo ) == 1 )
	    tf.die( "error while writing photon %d", nphot);
    }

    if ( bpipe_errno != BPNOERROR )
	tf.die( bpipe_strerror( bpipe_errno ) );

    bpipe_delete(bpipe);
}

